import os
import numpy as np
from astropy.cosmology import FlatLambdaCDM
from lenstronomy.Cosmo.lens_cosmo import LensCosmo
from hierarc.Sampling.mcmc_sampling import MCMCSampler
import corner
import matplotlib.pyplot as plt
from scipy.stats import norm, median_abs_deviation

__all__ = ["reorder_to_tdlmc", "pred_to_natural_gaussian", "CosmoConverter", "get_lognormal_stats", "get_lognormal_stats_naive", "get_normal_stats", "get_normal_stats_naive", "remove_outliers_from_lognormal", "combine_lenses", "gaussian_ll_pdf"]

MAD_to_sig = 1.0/norm.ppf(0.75) # 1.4826 built into scipy, so not used.

class DeltaFunction:
    def __init__(self, true_value=0.0):
        self.true_value = true_value
    def rvs(self, random_state=None):
        return self.true_value

def gaussian_ll_pdf(x, mu, sigma):
    """Evaluates the (unnormalized) log of the normal PDF at point x
    
    Parameters
    ----------
    x : float or array-like
        point at which to evaluate the log pdf
    mu : float or array-like
        mean of the normal on a linear scale
    sigma : float or array-like
        standard deviation of the normal on a linear scale
        
    """
    log_pdf = -0.5*(x - mu)**2.0/sigma**2.0 #- np.log(sigma) - 0.5*np.log(2.0*np.pi)
    return log_pdf

def reorder_to_tdlmc(img_array, increasing_dec_i, abcd_ordering_i):
    """Apply the permutation scheme for reordering the list of ra, dec, and time delays to conform to the order in the TDLMC challenge

    Parameters
    ----------
    img_array : array-like
        array of properties corresponding to the AGN images

    Returns
    -------
    array-like
        `img_array` reordered to the TDLMC order
    """
    #print(img_array, increasing_dec_i.shape, abcd_ordering_i.shape)
    img_array = np.array(img_array)[increasing_dec_i][abcd_ordering_i]
    return img_array

def pred_to_natural_gaussian(pred_mu, pred_cov_mat, shift, scale):
    """Convert the BNN-predicted multivariate Gaussian parameters into the natural space counterparts by reverse transformation

    Parameters
    ----------
    pred_mu : np.array of shape `[Y_dim,]`
    pred_cov_mat : np.array of shape `[Y_dim, Y_dim]`
    scale : np.array of shape `[Y_dim,]`
        vector by which the features were scaled, e.g. the training-set feature standard deviations
    shift : np.array of shape `[Y_dim,]`
        vector by which the features were shifted, e.g. the training-set feature means

    Note
    ----
    Derive it or go here: https://math.stackexchange.com/questions/332441/affine-transformation-applied-to-a-multivariate-gaussian-random-variable-what

    Returns
    -------
    mu : np.array of shape `[Y_dim,]`
        mu in natural space
    cov_mat : np.array of shape `[Y_dim, Y_dim]`
        covariance matrix in natural space

    """
    mu = pred_mu*scale + shift
    A = np.diagflat(scale)
    cov_mat = np.matmul(np.matmul(A, pred_cov_mat), A.T) # There is a better way to do this...
    return mu, cov_mat

class CosmoConverter:
    """Convert the time-delay distance to H0 and vice versa

    Note
    ----
    This was modified from lenstronomy.Cosmo.cosmo_solver to handle array types.

    """
    def __init__(self, z_lens, z_src, H0=70.0, Om0=0.3):
        self.cosmo_fiducial = FlatLambdaCDM(H0=H0, Om0=Om0) # arbitrary
        self.h0_fiducial = self.cosmo_fiducial.H0.value
        self.lens_cosmo = LensCosmo(z_lens=z_lens, z_source=z_src, cosmo=self.cosmo_fiducial)
        self.ddt_fiducial = self.lens_cosmo.ddt

    def get_H0(self, D_dt):
        H0 = self.h0_fiducial * self.ddt_fiducial / D_dt
        return H0

    def get_D_dt(self, H0):
        D_dt = self.h0_fiducial * self.ddt_fiducial / H0
        return D_dt

def get_lognormal_stats(all_samples):
    """Compute lognormal stats robustly, using median stats, assuming the samples are drawn from a lognormal distribution

    """
    is_nan_mask = np.logical_or(np.isnan(all_samples), ~np.isfinite(all_samples))
    samples = all_samples[~is_nan_mask]
    log_samples = np.log(samples)
    mu = np.median(log_samples)
    sig2 = median_abs_deviation(log_samples, axis=None, scale='normal')**2.0
    mode = np.exp(mu - sig2)
    std = ((np.exp(sig2) - 1.0)*(np.exp(2*mu - sig2)))**0.5
    stats = dict(
                 mu=mu,
                 sigma=sig2**0.5,
                 mode=mode,
                 std=std
                 )
    return stats

def get_lognormal_stats_naive(all_samples, all_weights=None):
    """Compute lognormal stats assuming the samples are drawn from a lognormal distribution

    """
    if all_weights is None:
        all_weights = np.ones_like(all_samples)
    is_nan_mask = np.logical_or(np.logical_or(np.isnan(all_weights), ~np.isfinite(all_weights)), np.isnan(all_samples))
    all_weights[~is_nan_mask] = all_weights[~is_nan_mask]/np.sum(all_weights[~is_nan_mask])
    samples = all_samples[~is_nan_mask]
    weights = all_weights[~is_nan_mask]
    n_samples = len(samples)
    log_samples = np.log(samples)
    mu = np.average(log_samples, weights=weights)
    sig2 = np.average((log_samples - mu)**2.0, weights=weights)*(n_samples/(n_samples - 1))
    mode = np.exp(mu - sig2)
    std = ((np.exp(sig2) - 1.0)*(np.exp(2*mu - sig2)))**0.5
    stats = dict(
                 mu=mu,
                 sigma=sig2**0.5,
                 mode=mode,
                 std=std
                 )
    return stats

def get_normal_stats(all_samples):
    is_nan_mask = np.logical_or(np.isnan(all_samples), ~np.isfinite(all_samples))
    samples = all_samples[~is_nan_mask]
    mean = np.median(samples)
    std = median_abs_deviation(samples, axis=None, scale='normal')
    stats = dict(
                 mean=mean,
                 std=std
                 )
    return stats

def get_normal_stats_naive(all_samples, all_weights):
    is_nan_mask = np.logical_or(np.logical_or(np.isnan(all_weights), ~np.isfinite(all_weights)), np.isnan(all_samples))
    all_weights[~is_nan_mask] = all_weights[~is_nan_mask]/np.sum(all_weights[~is_nan_mask])
    samples = all_samples[~is_nan_mask]
    weights = all_weights[~is_nan_mask]
    mean = np.average(samples, weights=weights)
    std = np.average((samples - mean)**2.0, weights=weights)**0.5
    #print(mean, std)
    stats = dict(
                 mean=mean,
                 std=std,
                 samples=samples,
                 weights=weights
                 )
    return stats

def remove_outliers_from_lognormal(data, level=3):
    """Remove extreme outliers corresponding to level-STD away from the mean
    
    Parameters
    ----------
    data : np.array
        data expected to follow a lognormal distribution

    """
    # Quantiles are preserved under monotonic transformations
    log_data = np.log(data)
    robust_mean = np.median(log_data)
    robust_std = median_abs_deviation(log_data, scale='normal')
    return data[abs(log_data - robust_mean) < level*robust_std]

def combine_lenses(likelihood_type, z_lens, z_src, true_Om0, samples_save_path=None, corner_save_path=None, n_run=100, n_burn=400, n_walkers=10, **posterior_parameters):
    """Combine lenses in the D_dt space

    Parameters
    ----------
    true_Om0 : float
        true Om0, not inferred
    likelihood_type : str
        'DdtGaussian', 'DdtLogNorm', 'DdtHistKDE' supported. 'DdtGaussian' must have 'ddt_mean', 'ddt_sigma'. 'DdtLogNorm' must have 'ddt_mu' and 'ddt_sigma'. 'DdtHistKDE' must have 'lens_ids' and 'samples_dir'.

    """
    n_test = len(z_lens)
    kwargs_posterior_list = []
    if likelihood_type in ['DdtLogNorm', 'DdtGaussian']:
        for i in range(n_test):
            kwargs_posterior = {'z_lens': z_lens[i], 'z_source': z_src[i],
                               'likelihood_type': likelihood_type}
            for param_name, param_value in posterior_parameters.items():
                kwargs_posterior.update({param_name: param_value[i]})
            kwargs_posterior_list.append(kwargs_posterior)
    elif likelihood_type == 'DdtHistKDE':
        lens_ids = posterior_parameters['lens_ids']
        samples_dir = posterior_parameters['samples_dir']
        binning_method = posterior_parameters['binning_method']
        for i, lens_i in enumerate(lens_ids):
            h0_dict_path = os.path.join(samples_dir, 'D_dt_dict_{:04d}.npy'.format(lens_i))
            h0_dict = np.load(h0_dict_path, allow_pickle=True).item() # TODO: Use context manager to prevent memory overload
            D_dt_samples = h0_dict['D_dt_samples']
            remove = np.isnan(D_dt_samples)
            D_dt_samples = D_dt_samples[~remove]
            #cosmo_converter = CosmoConverter(z_lens[i], z_src[i])
            #D_dt_samples = cosmo_converter.get_D_dt(H0_samples)
            kwargs_posterior = {'z_lens': z_lens[i], 'z_source': z_src[i], 
                                'ddt_samples': D_dt_samples, 'ddt_weights': None,
                               'likelihood_type': 'DdtHist', 'binning_method': binning_method}
            kwargs_posterior_list.append(kwargs_posterior)
    else:
        raise NotImplementedError("This likelihood type is not supported. Please choose from 'DdtGaussian', 'DdtLogNorm', and 'DdtHistKDE'.")

    kwargs_lower_cosmo = {'h0': 50.0}
    kwargs_lower_lens = {}
    kwargs_lower_kin = {}

    kwargs_upper_cosmo = {'h0': 90.0}
    kwargs_upper_lens = {}
    kwargs_upper_kin = {}

    kwargs_fixed_cosmo = {'om': true_Om0}
    kwargs_fixed_lens = {}
    kwargs_fixed_kin = {}

    kwargs_mean_start = {'kwargs_cosmo': {'h0': 70.0},
                         'kwargs_lens': {},
                         'kwargs_kin': {}}

    kwargs_sigma_start = {'kwargs_cosmo': {'h0': 10.0},
                         'kwargs_lens': {},
                         'kwargs_kin': {}}

    kwargs_bounds = {'kwargs_lower_cosmo': kwargs_lower_cosmo,
                'kwargs_lower_lens': kwargs_lower_lens,
                'kwargs_lower_kin': kwargs_lower_kin,
                'kwargs_upper_cosmo': kwargs_upper_cosmo,
                'kwargs_upper_lens': kwargs_upper_lens,
                'kwargs_upper_kin': kwargs_upper_kin,
                'kwargs_fixed_cosmo': kwargs_fixed_cosmo,
                'kwargs_fixed_lens': kwargs_fixed_lens,
                'kwargs_fixed_kin': kwargs_fixed_kin}

    cosmology = 'FLCDM'  # available models: 'FLCDM', "FwCDM", "w0waCDM", "oLCDM"
    mcmc_sampler = MCMCSampler(kwargs_likelihood_list=kwargs_posterior_list, 
                               cosmology=cosmology, 
                               kwargs_bounds=kwargs_bounds, 
                               ppn_sampling=False,
                               lambda_mst_sampling=False, 
                               lambda_mst_distribution='NONE', 
                               anisotropy_sampling=False,
                               kappa_ext_sampling=False, 
                               kappa_ext_distribution='NONE',
                               anisotropy_model='NONE', 
                               anisotropy_distribution='NONE', 
                               custom_prior=None, 
                               interpolate_cosmo=True, 
                               num_redshift_interp=100,
                               cosmo_fixed=None)

    mcmc_samples, log_prob_cosmo = mcmc_sampler.mcmc_emcee(n_walkers, n_run, n_burn, kwargs_mean_start, kwargs_sigma_start)
    if samples_save_path is not None:
        np.save(samples_save_path, mcmc_samples)

    if corner_save_path is not None:
        corner.corner(mcmc_samples, show_titles=True, labels=mcmc_sampler.param_names(latex_style=True))
        plt.show()
        plt.savefig(corner_save_path)
        plt.close()

    return mcmc_samples, log_prob_cosmo



import os
import glob
import numpy as np
import pandas as pd
import torch
from torch.utils.data import Dataset
import torchvision.transforms as transforms
from baobab import BaobabConfig
from baobab.data_augmentation.noise_torch import NoiseModelTorch
from baobab.sim_utils import add_g1g2_columns
from .data_utils import whiten_pixels, rescale_01, plus_1_log, whiten_Y_cols

__all__ = ['XYData']

class XYData(Dataset): # torch.utils.data.Dataset
    """Represents the XYData used to train or validate the BNN

    """
    def __init__(self, is_train, Y_cols, float_type, define_src_pos_wrt_lens, rescale_pixels, log_pixels, add_pixel_noise, eff_exposure_time, train_Y_mean=None, train_Y_std=None, train_baobab_cfg_path=None, val_baobab_cfg_path=None, for_cosmology=False, rescale_pixels_type='whiten_pixels'):
        """
        Parameters
        ----------
        dataset_dir : str or os.path object
            path to the directory containing the images and metadata
        data_cfg : dict or Dict
            copy of the `data` field of `BNNConfig`
        for_cosmology : bool
            whether the dataset will be used in cosmological inference 
            (in which case cosmology-related metadata will be stored)

        """
        #self.__dict__ = data_cfg.deepcopy()
        self.is_train = is_train
        if self.is_train:
            self.baobab_cfg = BaobabConfig.from_file(train_baobab_cfg_path)
        else:
            self.baobab_cfg = BaobabConfig.from_file(val_baobab_cfg_path)
        self.dataset_dir = self.baobab_cfg.out_dir
        if not self.is_train:
            if train_Y_mean is None or train_Y_std is None:
                raise ValueError("Mean and std of training set must be provided for whitening.")
        self.train_Y_mean = train_Y_mean
        self.train_Y_std = train_Y_std
        self.Y_cols = Y_cols
        self.float_type = float_type
        self.float_type_numpy = np.float64 if 'Double' in float_type else np.float32
        self.define_src_pos_wrt_lens = define_src_pos_wrt_lens
        self.rescale_pixels = rescale_pixels
        self.log_pixels = log_pixels
        self.add_pixel_noise = add_pixel_noise
        self.eff_exposure_time = eff_exposure_time
        self.bandpass_list = self.baobab_cfg.survey_info.bandpass_list
        self.for_cosmology = for_cosmology
        
        #################
        # Target labels #
        #################
        metadata_path = os.path.join(self.dataset_dir, 'metadata.csv')
        Y_df = pd.read_csv(metadata_path, index_col=False)
        if 'external_shear_gamma1' not in Y_df.columns: # assumes gamma_ext, psi_ext were sampled
            Y_df = add_g1g2_columns(Y_df)
        # Define source light position as offset from lens mass
        if self.define_src_pos_wrt_lens:
            Y_df['src_light_center_x'] -= Y_df['lens_mass_center_x']
            Y_df['src_light_center_y'] -= Y_df['lens_mass_center_y']
        train_Y_to_whiten = Y_df[self.Y_cols].values
        if self.is_train:
            self.train_Y_mean = np.mean(train_Y_to_whiten, axis=0, keepdims=True)
            self.train_Y_std = np.std(train_Y_to_whiten, axis=0, keepdims=True)
        # Store the unwhitened metadata
        if self.for_cosmology:
            self.Y_df = Y_df.copy()        
        # Number of predictive columns
        self.Y_dim = len(self.Y_cols)
        # Whiten the columns
        whiten_Y_cols(Y_df, self.train_Y_mean, self.train_Y_std, self.Y_cols)
        # Convert into array the columns required for training
        self.img_filenames = Y_df['img_filename'].values
        self.Y_array = Y_df[self.Y_cols].values.astype(self.float_type_numpy)
        # Free memory
        if not self.for_cosmology:
            del Y_df
        ################
        # Input images #
        ################
        # Set some metadata
        img_path = glob.glob(os.path.join(self.dataset_dir, '*.npy'))[0]
        img = np.load(img_path)
        self.X_dim = img.shape[0]

        # Rescale pixels, stack filters, and shift/scale pixels on the fly 
        if rescale_pixels_type == 'rescale_01':
            rescale = transforms.Lambda(rescale_01)
        else:
            rescale = transforms.Lambda(whiten_pixels)
        log = transforms.Lambda(plus_1_log)
        transforms_list = []
        if self.log_pixels:
            transforms_list.append(log)
        if self.rescale_pixels:
            transforms_list.append(rescale)
        if len(transforms_list) == 0:
            self.X_transform = lambda x: x
        else:
            self.X_transform = transforms.Compose(transforms_list)
        # Noise-related kwargs
        self.noise_kwargs = {}
        self.noiseless_exposure_time = {}
        self.noise_model = {}
        self.exposure_time_factor = np.ones([len(self.bandpass_list), 1, 1]) # for broadcasting
        for i, bp in enumerate(self.bandpass_list):
            survey_object = self.baobab_cfg.survey_object_dict[bp]
            # Dictionary of SingleBand kwargs
            self.noise_kwargs[bp] = survey_object.kwargs_single_band()
            # Factor of effective exptime relative to exptime of the noiseless images
            self.exposure_time_factor[i, :, :] = self.eff_exposure_time[bp]/self.noise_kwargs[bp]['exposure_time']
            if self.add_pixel_noise:
                self.noise_kwargs[bp].update(exposure_time=self.eff_exposure_time[bp])
                # Dictionary of noise models
                self.noise_model[bp] = NoiseModelTorch(**self.noise_kwargs[bp])

    def __getitem__(self, index):
        # Image X
        img_filename = self.img_filenames[index]
        img_path = os.path.join(self.dataset_dir, img_filename)
        img = np.load(img_path)
        img *= self.exposure_time_factor
        img = torch.as_tensor(img.astype(self.float_type_numpy)) # np array type must match with default tensor type
        if self.add_pixel_noise:
            for i, bp in enumerate(self.bandpass_list):
                img[i, :, :] += self.noise_model[bp].get_noise_map(img[i, :, :])
        img = self.X_transform(img)
        # Label Y
        Y_row = self.Y_array[index, :]
        Y_row = torch.as_tensor(Y_row)
        return img, Y_row

    def __len__(self):
        return self.Y_array.shape[0]


import os, sys
import importlib
import json
from addict import Dict

class TestConfig:
    """Nested dictionary representing the configuration for h0_inference

    """
    def __init__(self, user_cfg):
        """
        Parameters
        ----------
        user_cfg : dict or Dict
            user-defined configuration
        
        """
        self.__dict__ = Dict(user_cfg)

    @classmethod
    def from_file(cls, user_cfg_path):
        """Alternative constructor that accepts the path to the user-defined configuration python file

        Parameters
        ----------
        user_cfg_path : str or os.path object
            path to the user-defined configuration python file

        """
        dirname, filename = os.path.split(os.path.abspath(user_cfg_path))
        module_name, ext = os.path.splitext(filename)
        sys.path.append(dirname)
        if ext == '.py':
            #user_cfg_file = map(__import__, module_name)
            #user_cfg = getattr(user_cfg_file, 'cfg')
            user_cfg_script = importlib.import_module(module_name)
            user_cfg = getattr(user_cfg_script, 'cfg')
            return cls(user_cfg)
        elif ext == '.json':
            with open(user_cfg_path, 'r') as f:
                user_cfg_str = f.read()
            user_cfg = Dict(json.loads(user_cfg_str))
            return cls(user_cfg)
        else:
            raise NotImplementedError("This extension is not supported.")


import numpy as np
import pandas as pd
import torch
from lenstronomy.Sampling.parameters import Param
import baobab.sim_utils.metadata_utils as metadata_utils
import h0rton.losses

__all__ = ['get_lens_kwargs', 'get_ps_kwargs', 'get_ps_kwargs_src_plane', 'get_light_kwargs', 'get_special_kwargs', 'postprocess_mcmc_chain', "HybridBNNPenalty", "get_idx_for_params", "remove_parameters_from_pred", "split_component_param", "dict_to_array"]

# Conversion from param to BNN column naming
baobab_to_param = dict(zip(['lens_mass_theta_E', 'lens_mass_gamma', 'lens_mass_e1', 'lens_mass_e2', 'lens_mass_center_x', 'lens_mass_center_y', 'external_shear_gamma1', 'external_shear_gamma2', 'src_light_R_sersic', 'src_light_center_x', 'src_light_center_y', 'D_dt'],
                           ['theta_E_lens0', 'gamma_lens0', 'e1_lens0', 'e2_lens0', 'center_x_lens0', 'center_y_lens0', 'gamma1_lens1', 'gamma2_lens1', 'R_sersic_source_light0', 'ra_source', 'dec_source', 'D_dt']))
param_to_baobab = dict(zip(['theta_E_lens0', 'gamma_lens0', 'e1_lens0', 'e2_lens0', 'center_x_lens0', 'center_y_lens0', 'gamma1_lens1', 'gamma2_lens1', 'R_sersic_source_light0', 'ra_source', 'dec_source', 'D_dt'], 
                           ['lens_mass_theta_E', 'lens_mass_gamma', 'lens_mass_e1', 'lens_mass_e2', 'lens_mass_center_x', 'lens_mass_center_y', 'external_shear_gamma1', 'external_shear_gamma2', 'src_light_R_sersic', 'src_light_center_x', 'src_light_center_y', 'D_dt']))

def get_lens_kwargs(init_dict, null_spread=False):
    """Get the SPEMD, SHEAR kwargs with the provided init values and some conservative sigma, fixed, lower, and upper

    Note
    ----
    The sigma, fixed, lower, and upper kwargs are hardcoded.

    Parameters
    ----------
    init_dict : dict
        the init values for each of the parameters in SPEMD and SHEAR

    """
    eps = 1.e-7 # some small fractional value
    kwargs_init_lens = [{'theta_E': init_dict['lens_mass_theta_E'], 'gamma': init_dict['lens_mass_gamma'], 'center_x': init_dict['lens_mass_center_x'], 'center_y': init_dict['lens_mass_center_y'], 'e1': init_dict['lens_mass_e1'], 'e2': init_dict['lens_mass_e2']}, {'gamma1': init_dict['external_shear_gamma1'], 'gamma2': init_dict['external_shear_gamma2']}]
    if null_spread:
        kwargs_sigma_lens = [{k: eps*v for k, v in kwargs_init_lens[0].items()}, {k: eps*v for k, v in kwargs_init_lens[1].items()}]
        kwargs_fixed_lens = [{}, {'ra_0': init_dict['lens_mass_center_x'], 'dec_0': init_dict['lens_mass_center_y']}] # FIXME: fixing shear at the wrong position won't affect time delays but caution in case you add likelihoods affected by shear
    else:
        kwargs_sigma_lens = [{'theta_E': 0.05, 'e1': 0.05, 'e2': 0.05, 'gamma': 0.05, 'center_x': 0.02, 'center_y': 0.02}, {'gamma1': 0.05, 'gamma2': 0.05}]
        kwargs_fixed_lens = [{}, {'ra_0': 0.0, 'dec_0': 0.0}]
    kwargs_lower_lens = [{'theta_E': 0.01, 'e1': -1, 'e2': -1, 'gamma': 1, 'center_x': -10, 'center_y': -10}, {'gamma1': -1, 'gamma2': -1}]
    kwargs_upper_lens = [{'theta_E': 10, 'e1': 1, 'e2': 1, 'gamma': 4, 'center_x': 10, 'center_y': 10}, {'gamma1': 1, 'gamma2': 1,}]
    return [kwargs_init_lens, kwargs_sigma_lens, kwargs_fixed_lens, kwargs_lower_lens, kwargs_upper_lens]

def get_ps_kwargs(measured_img_ra, measured_img_dec, astrometry_sigma, hard_bound=30.0):
    """Get the point source kwargs for the image positions

    Parameters
    ----------
    measured_img_ra : np.array
        measured ra of the images
    measured_img_dec : np.array
        measured dec of the images
    astrometry_sigma : float
        astrometric uncertainty in arcsec
    hard_bound : float
        hard bound of the image positions around zero in arcsec

    Returns
    -------
    list of dict
        list of init, sigma, fixed, lower, and upper kwargs

    """
    n_img = len(measured_img_dec)
    ones = np.ones(n_img)
    kwargs_ps_init = [{'ra_image': measured_img_ra, 'dec_image': measured_img_dec}]
    kwargs_ps_sigma = [{'ra_image': astrometry_sigma*ones, 'dec_image': astrometry_sigma*ones}]
    fixed_ps = [{}] 
    kwargs_lower_ps = [{'ra_image': -hard_bound*ones, 'dec_image': -hard_bound*ones}]
    kwargs_upper_ps = [{'ra_image': hard_bound*ones, 'dec_image': hard_bound*ones}]
    return [kwargs_ps_init, kwargs_ps_sigma, fixed_ps, kwargs_lower_ps, kwargs_upper_ps]

def get_ps_kwargs_src_plane(init_dict, astrometry_sigma, hard_bound=5.0):
    """Get the point source kwargs for the source plane

    Parameters
    ----------
    measured_img_ra : np.array
        measured ra of the images
    measured_img_dec : np.array
        measured dec of the images
    astrometry_sigma : float
        astrometric uncertainty in arcsec
    hard_bound : float
        hard bound of the image positions around zero in arcsec

    Returns
    -------
    list of dict
        list of init, sigma, fixed, lower, and upper kwargs

    """
    kwargs_ps_init = [{'ra_source': init_dict['src_light_center_x'] + init_dict['lens_mass_center_x'], 'dec_source': init_dict['src_light_center_y'] + init_dict['lens_mass_center_y']}]
    kwargs_ps_sigma = [{'ra_source': astrometry_sigma, 'dec_source': astrometry_sigma}]
    fixed_ps = [{}] 
    kwargs_lower_ps = [{'ra_source': -hard_bound, 'dec_source': -hard_bound}]
    kwargs_upper_ps = [{'ra_source': hard_bound, 'dec_source': hard_bound}]
    return [kwargs_ps_init, kwargs_ps_sigma, fixed_ps, kwargs_lower_ps, kwargs_upper_ps]

def get_light_kwargs(init_R, null_spread=False):
    """Get the usual sigma, lower, upper, and fixed kwargs for a SERSIC_ELLIPSE light with only the R_sersic allowed to vary

    """
    eps = 1.e-7 # some small fractional value
    kwargs_light_init = [{'R_sersic': init_R}]
    if null_spread:
        kwargs_light_sigma = [{k: v*eps for k, v in kwargs_light_init[0].items()}]
    else:
        kwargs_light_sigma = [{'R_sersic': 0.05}]
    kwargs_light_fixed = [{'n_sersic': None, 'e1': None, 'e2': None, 'center_x': None, 'center_y': None}]
    kwargs_light_lower = [{'R_sersic': 0.01}]
    kwargs_light_upper = [{'R_sersic': 10.0}]
    return [kwargs_light_init, kwargs_light_sigma, kwargs_light_fixed, kwargs_light_lower, kwargs_light_upper]

def get_special_kwargs(n_img, astrometry_sigma, delta_pos_hard_bound=5.0, D_dt_init=5000.0, D_dt_sigma=1000.0, D_dt_lower=0.0, D_dt_upper=15000.0):
    """Get the point source kwargs for the image positions

    Parameters
    ----------
    measured_img_ra : np.array
        measured ra of the images
    measured_img_dec : np.array
        measured dec of the images
    astrometry_sigma : float
        astrometric uncertainty in arcsec
    hard_bound : float
        hard bound of the image positions around zero in arcsec

    Returns
    -------
    list of dict
        list of init, sigma, fixed, lower, and upper kwargs

    """
    zeros = np.zeros(n_img)
    ones = np.ones(n_img)
    kwargs_special_init = {'delta_x_image': zeros, 'delta_y_image': zeros, 'D_dt': D_dt_init}
    kwargs_special_sigma = {'delta_x_image': ones*astrometry_sigma, 'delta_y_image': ones*astrometry_sigma, 'D_dt': D_dt_sigma}
    fixed_special = {}
    kwargs_lower_special = {'delta_x_image': -ones*delta_pos_hard_bound, 'delta_y_image': -ones*delta_pos_hard_bound, 'D_dt': D_dt_lower}
    kwargs_upper_special = {'delta_x_image': ones*delta_pos_hard_bound, 'delta_y_image': ones*delta_pos_hard_bound, 'D_dt': D_dt_upper}
    return [kwargs_special_init, kwargs_special_sigma, fixed_special, kwargs_lower_special, kwargs_upper_special]

def postprocess_mcmc_chain(kwargs_result, samples, kwargs_model, fixed_lens_kwargs, fixed_ps_kwargs, fixed_src_light_kwargs, fixed_special_kwargs, kwargs_constraints, kwargs_fixed_lens_light=None, verbose=False, forward_modeling=False):
    """Postprocess the MCMC chain for making the chains consistent with the optimized lens model and converting parameters

    Returns
    -------
    pandas.DataFrame
        processed MCMC chain, where each row is a sample

    """
    param = Param(kwargs_model, fixed_lens_kwargs, kwargs_fixed_ps=fixed_ps_kwargs, kwargs_fixed_source=fixed_src_light_kwargs, kwargs_fixed_special=fixed_special_kwargs, kwargs_fixed_lens_light=kwargs_fixed_lens_light, kwargs_lens_init=kwargs_result['kwargs_lens'], **kwargs_constraints)
    if verbose:
        param.print_setting()
    n_samples = len(samples)
    processed = []
    for i in range(n_samples):
        kwargs = {}
        kwargs_out = param.args2kwargs(samples[i])
        kwargs_lens_out, kwargs_special_out, kwargs_ps_out, kwargs_source_out, kwargs_lens_light_out = kwargs_out['kwargs_lens'], kwargs_out['kwargs_special'], kwargs_out['kwargs_ps'], kwargs_out['kwargs_source'], kwargs_out['kwargs_lens_light']
        for k, v in kwargs_lens_out[0].items():
            kwargs['lens_mass_{:s}'.format(k)] = v
        for k, v in kwargs_lens_out[1].items():
            kwargs['external_shear_{:s}'.format(k)] = v
        if forward_modeling:
            for k, v in kwargs_source_out[0].items():
                kwargs['src_light_{:s}'.format(k)] = v
            for k, v in kwargs_lens_light_out[0].items():
                kwargs['lens_light_{:s}'.format(k)] = v
        else:
            kwargs['src_light_R_sersic'] = kwargs_source_out[0]['R_sersic']
        if 'ra_source' in kwargs_ps_out[0]:
            kwargs['src_light_center_x'] = kwargs_ps_out[0]['ra_source'] - kwargs_lens_out[0]['center_x']
            kwargs['src_light_center_y'] = kwargs_ps_out[0]['dec_source'] - kwargs_lens_out[0]['center_y']
        for k, v in kwargs_special_out.items():
            kwargs[k] = v
        processed.append(kwargs)
    processed_df = pd.DataFrame(processed)
    processed_df = metadata_utils.add_qphi_columns(processed_df)
    processed_df = metadata_utils.add_gamma_psi_ext_columns(processed_df)
    return processed_df

class HybridBNNPenalty:
    """Wrapper for subclasses of BaseGaussianNLL that allows MCMC methods to appropriately penalize parameter samples

    """
    def __init__(self, Y_cols, likelihood_class, mcmc_train_Y_mean, mcmc_train_Y_std, exclude_vel_disp, device):
        """
        Parameters
        ----------
        Y_dim : int
            number of parameters subject to penalty function
        likelihood_class : str
            name of subclass of BaseGaussianNLL to wrap around
        exclude_vel_disp : bool
            whether to add the NLL of velocity dispersion (not used)
        device : str

        """
        self.Y_cols = Y_cols
        self.Y_dim = len(self.Y_cols)
        self.mcmc_train_Y_mean = mcmc_train_Y_mean
        self.mcmc_train_Y_std = mcmc_train_Y_std
        self.constant_term = np.log(2*np.pi)*self.Y_dim*0.5
        #self.device = device
        self.exclude_vel_disp = exclude_vel_disp
        self.nll = getattr(h0rton.losses, '{:s}CPU'.format(likelihood_class))(Y_dim=self.Y_dim)

    def set_bnn_post_params(self, bnn_post_params):
        """Set BNN posterior parameters, which define the penaty function

        """
        self.bnn_post_params = bnn_post_params#.reshape(1, -1)
        self.n_dropout = bnn_post_params.shape[0]

    def evaluate(self, kwargs_lens, kwargs_source, kwargs_lens_light=None, kwargs_ps=None, kwargs_special=None, kwargs_extinction=None):
        #kwargs_lens[1]['ra_0'] = kwargs_lens[0]['center_x']
        #kwargs_lens[1]['dec_0'] = kwargs_lens[0]['center_y']
        to_eval = dict_to_array(self.Y_cols, kwargs_lens, kwargs_source, kwargs_lens_light, kwargs_ps) # shape [1, self.Y_dim]
        # Whiten the mcmc array
        to_eval = (to_eval - self.mcmc_train_Y_mean)/self.mcmc_train_Y_std
        #print(self.bnn_post_params[:, :self.Y_dim], to_eval)
        to_eval = np.repeat(to_eval, repeats=self.n_dropout, axis=0)
        ll = -self.nll(self.bnn_post_params, to_eval)
        return ll #+ self.constant_term

def get_idx_for_params(out_dim, Y_cols, params_to_remove, likelihood_class, debug=False):
    """Get columns corresponding to certain parameters from the BNN output

    Parameters
    ----------
    out_dim : int
    Y_cols : list of str
    
    Returns
    -------
    remove_idx : list
        indices of the columns removed in orig_pred
    remove_param_idx : list
        indices of the parameters in Y_cols

    """
    Y_dim = len(Y_cols)
    col_to_idx = dict(zip(Y_cols, range(Y_dim)))
    param_idx = np.array([col_to_idx[i] for i in params_to_remove]) # indices corresponding to primary mean
    if likelihood_class in ['FullRankGaussianNLL', 'DoubleGaussianNLL']:
        tril_idx = np.tril_indices(Y_dim)
        tril_idx_dim0 = tril_idx[0]
        tril_idx_dim1 = tril_idx[1]
        tril_len = len(tril_idx_dim0)
        tril_mask = np.logical_or(np.isin(tril_idx_dim0, param_idx), np.isin(tril_idx_dim1, param_idx)).nonzero()[0]
        idx_within_tril1 = list(Y_dim + tril_mask)
        idx_within_tril2 = list(2*Y_dim + tril_len + tril_mask)
        idx = list(param_idx) + idx_within_tril1 + list(Y_dim + tril_len + param_idx) + idx_within_tril2
        if debug:
            to_test = dict(
                           tril_mask=tril_mask,
                           idx_within_tril1=idx_within_tril1,
                           idx_within_tril2=idx_within_tril2,
                           param_idx=param_idx,
                           idx=idx,
                           )
            return to_test
    else: # tested for 'DoubleLowRankGaussianNLL':
        tiling_by_Y_dim = np.arange(out_dim//Y_dim)*Y_dim
        idx = [tile + i for tile in tiling_by_Y_dim for i in param_idx]
    param_idx = np.array(param_idx, dtype=int)
    idx = np.array(idx, dtype=int)
    return param_idx, idx 

def remove_parameters_from_pred(orig_pred, remove_idx, return_as_tensor=True, device='cpu'):
    """Remove columns corresponding to certain parameters from the BNN output

    Parameters
    ----------
    orig_pred : np.array of shape `[n_lenses, out_dim]`
        the BNN output
    orig_Y_cols : list of str
        the original list of Y columns
    params_to_remove : list of str
        list of colums to remove. They must belong to orig_Y_cols

    Returns
    -------
    new_pred : np.array of shape `[n_lenses, out_dim - len(params_to_remove)]`
    
    """
    new_pred = np.delete(orig_pred, remove_idx, axis=1)
    if return_as_tensor:
        new_pred = torch.as_tensor(new_pred, device=device)
    return new_pred

def split_component_param(string, sep='_', pos=2):
    """Split the component, e.g. lens_mass, from the parameter, e.g. center_x, from the paramter names under the Baobab convention

    Parameters
    ----------
    string : str
        the Baobab parameter name (column name)
    sep : str
        separation character between the component and the parameter. Default: '_' (Baobab convention)
    pos : int
        position of the component when split by the separation character. Default: 2 (Baobab convention)

    Returns
    -------
    tuple of str
        component, parameter

    """
    substring_list = string.split(sep)
    return sep.join(substring_list[:pos]), sep.join(substring_list[pos:])

def dict_to_array(Y_cols, kwargs_lens, kwargs_source, kwargs_lens_light=None, kwargs_ps=None,):
    """Reformat kwargs into np array. Used to feed the current iteration of MCMC kwargs into the BNN posterior evaluation.

    """
    return_array = np.ones((len(Y_cols)))
    for i, col_name in enumerate(Y_cols):
        component, param = split_component_param(col_name, '_', 2)
        if component == 'lens_mass':
            return_array[i] = kwargs_lens[0][param]
        elif component == 'external_shear':
            return_array[i] = kwargs_lens[1][param]
        elif component == 'src_light':
            if param == 'center_x':
                return_array[i] = kwargs_ps[0]['ra_source'] - kwargs_lens[0]['center_x']
            elif param == 'center_y':
                return_array[i] = kwargs_ps[0]['dec_source'] - kwargs_lens[0]['center_y']
            else:
                return_array[i] = kwargs_source[0][param]
        #elif component == 'lens_light':
        #    return_array[i] = kwargs_lens_light[0][param]
        else:
            # Ignore kwargs_ps since image position likelihood is separate.
            raise ValueError("Component doesn't exist.")
    return return_array.reshape(1, -1)

def reorder_to_param_class(bnn_Y_cols, param_class_Y_cols, bnn_array, D_dt_array):
    """Reorder an array with the given axis ordered according to the BNN's bnn_Y_cols to the Lenstronomy Param's param_class_Y_cols convention

    Parameters
    ----------
    bnn_Y_cols : list of str
    param_class_Y_cols : list of str
    bnn_array : np.array 
    D_dt_array : np.array 
    axis : int
    
    """
    baobab_cols = bnn_Y_cols + ['D_dt']
    baobab_col_to_idx = dict(zip(baobab_cols, range(len(baobab_cols))))
    #param_col_to_idx = dict(zip(param_class_Y_cols, range(len(param_class_Y_cols))))
    # Store the absolute source position instead
    bnn_array[:, :, baobab_col_to_idx['src_light_center_x']] += bnn_array[:, :, baobab_col_to_idx['lens_mass_center_x']]
    bnn_array[:, :, baobab_col_to_idx['src_light_center_y']] += bnn_array[:, :, baobab_col_to_idx['lens_mass_center_y']]
    baobab_array = np.concatenate([bnn_array, D_dt_array], axis=-1) # [n_lenses, n_samples, bnn_Y_dim + 1]
    ordering = [baobab_col_to_idx[param_to_baobab[param_col]] for param_col in param_class_Y_cols]
    mcmc_array = baobab_array[:, :, ordering] # [n_lenses, n_samples, len(param_class_Y_cols)]
    return mcmc_array


from abc import ABC, abstractmethod
import numpy as np
import torch
__all__ = ['BaseGaussianNLL', 'DiagonalGaussianNLL', 'LowRankGaussianNLL', 'DoubleLowRankGaussianNLL', 'FullRankGaussianNLL', 'DoubleGaussianNLL']

log_2_pi = 1.8378770664093453
log_2 = 0.6931471805599453

class BaseGaussianNLL(ABC):
    """Abstract base class to represent the Gaussian negative log likelihood (NLL).

    Gaussian NLLs or mixtures thereof with various forms of the covariance matrix inherit from this class.

    """
    def __init__(self, Y_dim, device):
        """
        Parameters
        ----------
        Y_dim : int
            number of parameters to predict
        device : torch.device object

        """
        self.Y_dim = Y_dim
        self.device = device
        self.sigmoid = torch.nn.Sigmoid()
        self.logsigmoid = torch.nn.LogSigmoid()

    @abstractmethod
    def slice(self, pred):
        """Slice the raw network prediction into meaningful Gaussian parameters

        Parameters
        ----------
        pred : torch.Tensor of shape `[batch_size, self.Y_dim]`
            the network prediction

        """
        return NotImplemented

    @abstractmethod
    def __call__(self, pred, target):
        """Evaluate the NLL. Must be overridden by subclasses.

        Parameters
        ----------
        pred : torch.Tensor
            raw network output for the predictions
        target : torch.Tensor
            Y labels

        """
        return NotImplemented

    def nll_diagonal(self, target, mu, logvar):
        """Evaluate the NLL for single Gaussian with diagonal covariance matrix

        Parameters
        ----------
        target : torch.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior
        logvar : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the log of the diagonal elements of the covariance matrix

        Returns
        -------
        torch.Tensor of shape
            NLL values

        """
        precision = torch.exp(-logvar)
        # Loss kernel
        loss = precision * (target - mu)**2.0 + logvar
        # Restore prefactors
        loss += np.log(2.0*np.pi)
        loss *= 0.5
        return torch.mean(torch.sum(loss, dim=1), dim=0)

    def nll_low_rank(self, target, mu, logvar, F, reduce=True):
        """Evaluate the NLL for a single Gaussian with a full but low-rank plus diagonal covariance matrix

        Parameters
        ----------
        target : torch.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior
        logvar : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the log of the diagonal elements of the covariance matrix
        F : torch.Tensor of shape [batch_size, rank*Y_dim]
            network prediction of the low rank portion of the covariance matrix
        reduce : bool
            whether to take the mean across the batch

        Returns
        -------
        torch.Tensor of shape [batch_size,]
            NLL values

        """
        # 1/(Y_dim - 1) * (sq_mahalanobis + log(det of \Sigma))
        batch_size, _ = target.shape # self.Y_dim = Y_dim - 1
        rank = 2 # FIXME: hardcoded for rank 2
        F = F.reshape([batch_size, self.Y_dim, rank]) 
        inv_var = torch.exp(-logvar) # [batch_size, self.Y_dim]
        diag_inv_var = torch.diag_embed(inv_var).float()  # [batch_size, self.Y_dim, self.Y_dim]
        diag_prod = F**2.0 * inv_var.reshape([batch_size, self.Y_dim, 1]) # [batch_size, self.Y_dim, rank] after broadcasting
        off_diag_prod = torch.prod(F, dim=2)*inv_var # [batch_size, self.Y_dim]
        #batchdiag = torch.diag_embed(torch.exp(logvar)) # [batch_size, self.Y_dim, self.Y_dim]
        #batch_eye = torch.eye(rank).reshape(1, rank, rank).repeat(batch_size, 1, 1) # [batch_size, rank, rank]
        # (25), (26) in Miller et al 2016
        log_det = torch.sum(logvar, dim=1) # [batch_size]
        M00 = torch.sum(diag_prod[:, :, 0], dim=1) + 1.0 # [batch_size]
        M11 = torch.sum(diag_prod[:, :, 1], dim=1) + 1.0 # [batch_size]
        M12 = torch.sum(off_diag_prod, dim=1) # [batch_size]
        det_M = M00*M11 - M12**2.0 # [batch_size]
        log_det += torch.log(det_M) # [batch_size,]

        inv_M = torch.ones([batch_size, rank, rank], device=self.device, dtype=None)
        inv_M[:, 0, 0] = M11
        inv_M[:, 1, 1] = M00
        inv_M[:, 1, 0] = -M12
        inv_M[:, 0, 1] = -M12
        inv_M /= det_M.reshape(batch_size, 1, 1)

        # (27) in Miller et al 2016
        inv_cov = diag_inv_var - torch.bmm(torch.bmm(torch.bmm(torch.bmm(diag_inv_var, F), inv_M), torch.transpose(F, 1, 2)), diag_inv_var) # [batch_size, self.Y_dim, self.Y_dim]
        sq_mahalanobis = torch.bmm(torch.bmm((mu - target).reshape(batch_size, 1, self.Y_dim), inv_cov), (mu - target).reshape(batch_size, self.Y_dim, 1)).reshape(-1) # [batch_size,]
        
        # Loss kernel
        loss = sq_mahalanobis + log_det
        # Restore prefactors
        loss += self.Y_dim*np.log(2.0*np.pi)
        loss *= 0.5

        if reduce==True:
            return torch.mean(loss, dim=0) # float
        else:
            return loss # [batch_size,]

    def nll_mixture_low_rank(self, target, mu, logvar, F, mu2, logvar2, F2, alpha):
        """Evaluate the NLL for a single Gaussian with a full but low-rank plus diagonal covariance matrix
        Parameters
        ----------
        target : torch.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior for the first Gaussian
        logvar : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the log of the diagonal elements of the covariance matrix for the first Gaussian
        F : torch.Tensor of shape [batch_size, rank*Y_dim]
            network prediction of the low rank portion of the covariance matrix for the first Gaussian
        mu2 : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior for the second Gaussian
        logvar2 : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the log of the diagonal elements of the covariance matrix for the second Gaussian
        F2 : torch.Tensor of shape [batch_size, rank*Y_dim]
            network prediction of the low rank portion of the covariance matrix for the second Gaussian
        alpha : torch.Tensor of shape [batch_size, 1]
            network prediction of the logit of twice the weight on the second Gaussian 
        reduce : bool
            whether to take the mean across the batch
        Note
        ----
        The weight on the second Gaussian is required to be less than 0.5, to make the two Gaussians well-defined.
        Returns
        -------
        torch.Tensor of shape [batch_size,]
            NLL values
        """
        batch_size, _ = target.shape
        alpha = alpha.reshape(-1)
        #log_ll[:, 0] = torch.log1p(-0.5*self.sigmoid(alpha)) - self.nll_low_rank(target, mu, logvar, F=F, reduce=False) # [batch_size]
        # torch.log(torch.tensor([0.5], device=self.device)).double()
        #log_ll[:, 1] = -log_2 + self.logsigmoid(alpha) - self.nll_low_rank(target, mu2, logvar2, F=F2, reduce=False) # [batch_size], 0.6931471 = np.log(2)
        log_w1p1 = torch.log1p(2.0*torch.exp(-alpha)) - log_2 - torch.log1p(torch.exp(-alpha)) - self.nll_low_rank(target, mu, logvar, F=F, reduce=False) # [batch_size]
        log_w2p2 = -log_2 + self.logsigmoid(alpha) - self.nll_low_rank(target, mu2, logvar2, F=F2, reduce=False) # [batch_size], 0.6931471 = np.log(2)
        stacked = torch.stack([log_w1p1, log_w2p2], dim=1)
        log_nll = -torch.logsumexp(stacked, dim=1)
        return torch.mean(log_nll)

    def nll_full_rank(self, target, mu, tril_elements, reduce=True):
        """Evaluate the NLL for a single Gaussian with a full-rank covariance matrix

        Parameters
        ----------
        target : torch.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior
        tril_elements : torch.Tensor of shape [batch_size, Y_dim*(Y_dim + 1)//2]
        reduce : bool
            whether to take the mean across the batch

        Returns
        -------
        torch.Tensor of shape [batch_size,]
            NLL values

        """
        batch_size, _ = target.shape
        tril = torch.zeros([batch_size, self.Y_dim, self.Y_dim], device=self.device, dtype=None)
        tril[:, self.tril_idx[0], self.tril_idx[1]] = tril_elements
        log_diag_tril = torch.diagonal(tril, offset=0, dim1=1, dim2=2) # [batch_size, Y_dim]
        logdet_term = -torch.sum(log_diag_tril, dim=1) # [batch_size,]
        tril[:, torch.eye(self.Y_dim, dtype=bool)] = torch.exp(log_diag_tril)
        prec_mat = torch.bmm(tril, torch.transpose(tril, 1, 2)) # [batch_size, Y_dim, Y_dim]
        y_diff = mu - target # [batch_size, Y_dim]
        mahalanobis_term = 0.5*torch.sum(
            y_diff*torch.sum(prec_mat*y_diff.unsqueeze(-1), dim=-2), dim=-1) # [batch_size,]
        loss = logdet_term + mahalanobis_term + 0.5*self.Y_dim*log_2_pi
        if reduce:
            return torch.mean(loss, dim=0) # float
        else:
            return loss # [batch_size,]

    def nll_mixture(self, target, mu, tril_elements, mu2, tril_elements2, alpha):
        """Evaluate the NLL for a single Gaussian with a full but low-rank plus diagonal covariance matrix

        Parameters
        ----------
        target : torch.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior for the first Gaussian
        tril_elements : torch.Tensor of shape [batch_size, self.tril_len]
            network prediction of the elements in the precision matrix
        mu2 : torch.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior for the second Gaussian
        tril_elements2 : torch.Tensor of shape [batch_size, self.tril_len]
            network prediction of the elements in the precision matrix for the second Gaussian
        alpha : torch.Tensor of shape [batch_size, 1]
            network prediction of the logit of twice the weight on the second Gaussian 

        Note
        ----
        The weight on the second Gaussian is required to be less than 0.5, to make the two Gaussians well-defined.

        Returns
        -------
        torch.Tensor of shape [batch_size,]
            NLL values

        """
        batch_size, _ = target.shape
        alpha = alpha.reshape(-1)
        #log_ll[:, 0] = torch.log1p(-0.5*self.sigmoid(alpha)) - self.nll_full_rank(target, mu, tril_elements, reduce=False) # [batch_size]
        #log_ll[:, 1] = -log_2 + self.logsigmoid(alpha) - self.nll_full_rank(target, mu2, tril_elements2, reduce=False) # [batch_size], 0.6931471 = np.log(2)
        log_w1p1 = torch.log1p(2.0*torch.exp(-alpha)) - log_2 - torch.log1p(torch.exp(-alpha)) - self.nll_full_rank(target, mu, tril_elements, reduce=False) # [batch_size]
        log_w2p2 = -log_2 + self.logsigmoid(alpha) - self.nll_full_rank(target, mu2, tril_elements2, reduce=False) # [batch_size]
        stacked = torch.stack([log_w1p1, log_w2p2], dim=1)
        log_nll = -torch.logsumexp(stacked, dim=1)
        return torch.mean(log_nll)

class DiagonalGaussianNLL(BaseGaussianNLL):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    posterior_name = 'DiagonalGaussianBNNPosterior'

    def __init__(self, Y_dim, device):
        super(DiagonalGaussianNLL, self).__init__(Y_dim, device)
        self.out_dim = Y_dim*2

    def __call__(self, pred, target):
        return self.nll_diagonal(target, *self.slice(pred))

    def slice(self, pred):
        d = self.Y_dim # for readability
        return torch.split(pred, [d, d], dim=1)

class LowRankGaussianNLL(BaseGaussianNLL):
    """The negative log likelihood (NLL) for a single Gaussian with a full but constrained as low-rank plus diagonal covariance matrix
        
    Only rank 2 is currently supported. `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    posterior_name = 'LowRankGaussianBNNPosterior'

    def __init__(self, Y_dim, device):
        super(LowRankGaussianNLL, self).__init__(Y_dim, device)
        self.out_dim = Y_dim*4

    def __call__(self, pred, target):
        return self.nll_low_rank(target, *self.slice(pred), reduce=True)

    def slice(self, pred):
        d = self.Y_dim # for readability
        return torch.split(pred, [d, d, 2*d], dim=1)

class FullRankGaussianNLL(BaseGaussianNLL):
    """The negative log likelihood (NLL) for a single Gaussian with a full-rank covariance matrix
        
    See `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    posterior_name = 'FullRankGaussianBNNPosterior'

    def __init__(self, Y_dim, device):
        super(FullRankGaussianNLL, self).__init__(Y_dim, device)
        self.tril_idx = torch.tril_indices(self.Y_dim, self.Y_dim, offset=0, device=device) # lower-triangular indices
        self.tril_len = len(self.tril_idx[0])
        self.out_dim = self.Y_dim + self.Y_dim*(self.Y_dim + 1)//2

    def __call__(self, pred, target):
        return self.nll_full_rank(target, *self.slice(pred), reduce=True)

    def slice(self, pred):
        d = self.Y_dim # for readability
        return torch.split(pred, [d, self.tril_len], dim=1)

class DoubleLowRankGaussianNLL(BaseGaussianNLL):
    """The negative log likelihood (NLL) for a mixture of two Gaussians, each with a full but constrained as low-rank plus diagonal covariance 
        
    Only rank 2 is currently supported. `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    posterior_name = 'DoubleLowRankGaussianBNNPosterior'

    def __init__(self, Y_dim, device):
        super(DoubleLowRankGaussianNLL, self).__init__(Y_dim, device)
        self.out_dim = Y_dim*8 + 1

    def __call__(self, pred, target):
        return self.nll_mixture_low_rank(target, *self.slice(pred))

    def slice(self, pred):
        d = self.Y_dim # for readability
        #mu, logvar, F, mu2, logvar2, F2, alpha
        return torch.split(pred, [d, d, 2*d, d, d, 2*d, 1], dim=1)

class DoubleGaussianNLL(BaseGaussianNLL):
    """The negative log likelihood (NLL) for a mixture of two Gaussians, each with a full but constrained as low-rank plus diagonal covariance 
        
    Only rank 2 is currently supported. `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    posterior_name = 'DoubleGaussianBNNPosterior'

    def __init__(self, Y_dim, device):
        super(DoubleGaussianNLL, self).__init__(Y_dim, device)
        self.tril_idx = torch.tril_indices(self.Y_dim, self.Y_dim, offset=0, device=device) # lower-triangular indices
        self.tril_len = len(self.tril_idx[0])
        self.out_dim = self.Y_dim**2 + 3*self.Y_dim + 1

    def __call__(self, pred, target):
        return self.nll_mixture(target, *self.slice(pred))

    def slice(self, pred):
        d = self.Y_dim # for readability
        return torch.split(pred, [d, self.tril_len, d, self.tril_len, 1], dim=1)


from abc import ABC, abstractmethod
import random
import numpy as np
from scipy.stats import multivariate_normal
__all__ = ['BaseGaussianBNNPosteriorCPU', 'DiagonalGaussianBNNPosteriorCPU', 'LowRankGaussianBNNPosteriorCPU', 'DoubleLowRankGaussianBNNPosteriorCPU', 'FullRankGaussianBNNPosteriorCPU', 'DoubleGaussianBNNPosteriorCPU', 'sigmoid', 'logsigmoid']

def sigmoid(x):
    return np.where(x >= 0, 
                    1 / (1 + np.exp(-x)), 
                    np.exp(x) / (1 + np.exp(x)))

def logsigmoid(x):
    return np.where(x >= 0, 
                    -np.log1p(np.exp(-x)), 
                    x - np.log1p(np.exp(x)))

class BaseGaussianBNNPosteriorCPU(ABC):
    """Abstract base class to represent the Gaussian BNN posterior

    Gaussian posteriors or mixtures thereof with various forms of the covariance matrix inherit from this class.

    """
    def __init__(self, Y_dim, Y_mean=None, Y_std=None):
        """
        Parameters
        ----------
        pred : torch.Tensor of shape `[1, out_dim]` or `[out_dim,]`
            raw network output for the predictions
        Y_dim : int
            number of parameters to predict
        Y_mean : list
            mean values for the original values of `whitened_Y_cols`
        Y_std : list
            std values for the original values of `whitened_Y_cols`
        device : torch.device object

        """
        self.Y_dim = Y_dim
        self.Y_mean = Y_mean.reshape(1, -1)
        self.Y_std = Y_std.reshape(1, -1)
        self.sigmoid = sigmoid
        self.logsigmoid = logsigmoid

    def seed_samples(self, sample_seed):
        """Seed the sampling for reproducibility

        Parameters
        ----------
        sample_seed : int

        """
        np.random.seed(sample_seed)
        random.seed(sample_seed)

    @abstractmethod
    def sample(self, n_samples, sample_seed=None):
        """Sample from the Gaussian posterior. Must be overridden by subclasses.

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        sample_seed : int
            seed for the samples. Default: None

        Returns
        -------
        np.array of shape `[n_samples, self.Y_dim]`
            samples

        """
        return NotImplemented

    @abstractmethod
    def get_hpd_interval(self):
        """Get the highest posterior density (HPD) interval

        """
        return NotImplemented

    def transform_back_mu(self, array):
        """Transform back, i.e. unwhiten, the tensor of central values

        Parameters
        ----------
        array : np.array of shape `[batch_size, Y_dim]`

        Returns
        -------
        torch.tensor of shape `[batch_size, Y_dim]`
            the original tensor

        """
        array = np.expand_dims(array, axis=1)
        array = self.unwhiten_back(array)
        return array.squeeze()

    def unwhiten_back(self, sample):
        """Scale and shift back to the unwhitened state

        Parameters
        ----------
        pred : torch.Tensor
            network prediction of shape `[batch_size, n_samples, self.Y_dim]`

        Returns
        -------
        torch.Tensor
            the unwhitened pred
        
        """
        sample = sample*np.expand_dims(self.Y_std, 1) + np.expand_dims(self.Y_mean, 1)
        return sample

    def sample_low_rank(self, n_samples, mu, logvar, F):
        """Sample from a single Gaussian posterior with a full but low-rank plus diagonal covariance matrix

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        mu : torch.Tensor of shape `[self.batch_size, self.Y_dim]`
            network prediction of the mu (mean parameter) of the BNN posterior
        logvar : torch.Tensor of shape `[self.batch_size, self.Y_dim]`
            network prediction of the log of the diagonal elements of the covariance matrix
        F : torch.Tensor of shape `[self.batch_size, self.Y_dim, self.rank]`
            network prediction of the low rank portion of the covariance matrix

        Returns
        -------
        np.array of shape `[self.batch_size, n_samples, self.Y_dim]`
            samples

        """
        #F = torch.unsqueeze(F, dim=1).repeat(1, n_samples, 1, 1) # [self.batch_size, n_samples, self.Y_dim, self.rank]
        F = np.repeat(F, repeats=n_samples, axis=0) # [self.batch_size*n_samples, self.Y_dim, self.rank]
        mu = np.repeat(mu, repeats=n_samples, axis=0) # [self.batch_size*n_samples, self.Y_dim]
        logvar = np.repeat(logvar, repeats=n_samples, axis=0) # [self.batch_size*n_samples, self.Y_dim]
        eps_low_rank = np.random.randn(self.batch_size*n_samples, self.rank, 1)
        eps_diag = np.random.randn(self.batch_size*n_samples, self.Y_dim)
        half_var = np.exp(0.5*logvar) # [self.batch_size*n_samples, self.Y_dim]
        samples = np.matmul(F, eps_low_rank).squeeze() + mu + half_var*eps_diag # [self.batch_size*n_samples, self.Y_dim]
        samples = samples.reshape(self.batch_size, n_samples, self.Y_dim)
        #samples = samples.transpose((1, 0, 2)) # [self.batch_size, n_samples, self.Y_dim]
        return samples
    
    def sample_full_rank(self, n_samples, mu, tril_elements):
        """Sample from a single Gaussian posterior with a full-rank covariance matrix

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        mu : torch.Tensor of shape `[self.batch_size, self.Y_dim]`
            network prediction of the mu (mean parameter) of the BNN posterior
        tril_elements : torch.Tensor of shape `[self.batch_size, tril_len]`
            network prediction of lower-triangular matrix in the log-Cholesky decomposition of the precision matrix

        Returns
        -------
        np.array of shape `[self.batch_size, n_samples, self.Y_dim]`
            samples

        """
        samples = np.zeros([self.batch_size, n_samples, self.Y_dim])
        for b in range(self.batch_size):
            tril = np.zeros([self.Y_dim, self.Y_dim])
            tril[self.tril_idx[0], self.tril_idx[1]] = tril_elements[b, :]
            log_diag_tril = np.diagonal(tril, offset=0, axis1=0, axis2=1)
            tril[np.eye(self.Y_dim).astype(bool)] = np.exp(log_diag_tril)
            prec_mat = np.dot(tril, tril.T) # [Y_dim, Y_dim]
            cov_mat = np.linalg.inv(prec_mat)
            sample_b = multivariate_normal.rvs(mean=mu[b, :], cov=cov_mat, size=[n_samples,])
            samples[b, :, :] = sample_b
        samples = self.unwhiten_back(samples)
        return samples

class DiagonalGaussianBNNPosteriorCPU(BaseGaussianBNNPosteriorCPU):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    def __init__(self, Y_dim, Y_mean=None, Y_std=None):
        super(DiagonalGaussianBNNPosteriorCPU, self).__init__(Y_dim, Y_mean, Y_std)
        self.out_dim = self.Y_dim*2

    def set_sliced_pred(self, pred):
        d = self.Y_dim # for readability
        #pred = pred.cpu().numpy()
        self.batch_size = pred.shape[0]
        self.mu = pred[:, :d]
        self.logvar = pred[:, d:]
        self.cov_diag = np.exp(self.logvar)
        #F_tran_F = np.matmul(normal.F, np.swapaxes(normal.F, 1, 2))
        #cov_mat = np.apply_along_axis(np.diag, -1, np.exp(normal.logvar)) + F_tran_F
        #cov_diag = np.exp(normal.logvar) + np.diagonal(F_tran_F, axis1=1, axis2=2)
        #assert np.array_equal(cov_mat.shape, [batch_size, self.Y_dim, self.Y_dim])
        #assert np.array_equal(cov_diag.shape, [batch_size, self.Y_dim])
        #np.apply_along_axis(np.diag, -1, np.exp(logvar)) # for diagonal

    def sample(self, n_samples, sample_seed):
        """Sample from a Gaussian posterior with diagonal covariance matrix

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        sample_seed : int
            seed for the samples. Default: None

        Returns
        -------
        np.array of shape `[n_samples, self.Y_dim]`
            samples

        """
        self.seed_samples(sample_seed)
        eps = np.random.randn(self.batch_size, n_samples, self.Y_dim)
        samples = eps*np.exp(0.5*np.expand_dims(self.logvar, 1)) + np.expand_dims(self.mu, 1)
        samples = self.unwhiten_back(samples)
        return samples

    def get_hpd_interval(self):
        return NotImplementedError

class LowRankGaussianBNNPosteriorCPU(BaseGaussianBNNPosteriorCPU):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    def __init__(self, Y_dim, Y_mean=None, Y_std=None):
        super(LowRankGaussianBNNPosteriorCPU, self).__init__(Y_dim, Y_mean, Y_std)
        self.out_dim = self.Y_dim*4
        self.rank = 2 # FIXME: hardcoded

    def set_sliced_pred(self, pred):
        d = self.Y_dim # for readability
        #pred = pred.cpu().numpy()
        self.batch_size = pred.shape[0]
        self.mu = pred[:, :d]
        self.logvar = pred[:, d:2*d]
        self.F = pred[:, 2*d:].reshape([self.batch_size, self.Y_dim, self.rank])
        #F_tran_F = np.matmul(self.F, np.swapaxes(self.F, 1, 2))
        #self.cov_diag = np.exp(self.logvar) + np.diagonal(F_tran_F, axis1=1, axis2=2)

    def sample(self, n_samples, sample_seed):
        self.seed_samples(sample_seed)
        return self.sample_low_rank(n_samples, self.mu, self.logvar, self.F)

    def get_hpd_interval(self):
        return NotImplementedError

class DoubleLowRankGaussianBNNPosteriorCPU(BaseGaussianBNNPosteriorCPU):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    def __init__(self, Y_dim, Y_mean=None, Y_std=None):
        super(DoubleLowRankGaussianBNNPosteriorCPU, self).__init__(Y_dim, Y_mean, Y_std)
        self.out_dim = self.Y_dim*8 + 1
        self.rank = 2 # FIXME: hardcoded

    def set_sliced_pred(self, pred):
        d = self.Y_dim # for readability
        #pred = pred.cpu().numpy()
        self.w2 = 0.5*self.sigmoid(pred[:, -1].reshape(-1, 1))
        self.batch_size = pred.shape[0]
        self.mu = pred[:, :d]
        self.logvar = pred[:, d:2*d]
        self.F = pred[:, 2*d:4*d].reshape([self.batch_size, self.Y_dim, self.rank])
        #F_tran_F = np.matmul(self.F, np.swapaxes(self.F, 1, 2))
        #self.cov_diag = np.exp(self.logvar) + np.diagonal(F_tran_F, axis1=1, axis2=2)
        self.mu2 = pred[:, 4*d:5*d]
        self.logvar2 = pred[:, 5*d:6*d]
        self.F2 = pred[:, 6*d:8*d].reshape([self.batch_size, self.Y_dim, self.rank])
        #F_tran_F2 = np.matmul(self.F2, np.swapaxes(self.F2, 1, 2))
        #self.cov_diag2 = np.exp(self.logvar2) + np.diagonal(F_tran_F2, axis1=1, axis2=2)
        
        
    def sample(self, n_samples, sample_seed):
        """Sample from a mixture of two Gaussians, each with a full but constrained as low-rank plus diagonal covariance

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        sample_seed : int
            seed for the samples. Default: None

        Returns
        -------
        np.array of shape `[self.batch_size, n_samples, self.Y_dim]`
            samples

        """
        self.seed_samples(sample_seed)
        samples = np.zeros([self.batch_size, n_samples, self.Y_dim])
        # Determine first vs. second Gaussian
        unif2 = np.random.rand(self.batch_size, n_samples)
        second_gaussian = (self.w2 > unif2)
        # Sample from second Gaussian
        samples2 = self.sample_low_rank(n_samples, self.mu2, self.logvar2, self.F2)
        samples[second_gaussian, :] = samples2[second_gaussian, :]
        # Sample from first Gaussian
        samples1 = self.sample_low_rank(n_samples, self.mu, self.logvar, self.F)
        samples[~second_gaussian, :] = samples1[~second_gaussian, :]
        return samples

    def get_hpd_interval(self):
        return NotImplementedError

class FullRankGaussianBNNPosteriorCPU(BaseGaussianBNNPosteriorCPU):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    def __init__(self, Y_dim, Y_mean=None, Y_std=None):
        super(FullRankGaussianBNNPosteriorCPU, self).__init__(Y_dim, Y_mean, Y_std)
        self.tril_idx = np.tril_indices(self.Y_dim) # lower-triangular indices
        self.tril_len = len(self.tril_idx[0])
        self.out_dim = self.Y_dim + self.Y_dim*(self.Y_dim + 1)//2

    def set_sliced_pred(self, pred):
        d = self.Y_dim # for readability
        #pred = pred.cpu().numpy()
        self.batch_size = pred.shape[0]
        self.mu = pred[:, :d]
        self.tril_elements = pred[:, d:self.out_dim]

    def sample(self, n_samples, sample_seed):
        self.seed_samples(sample_seed)
        return self.sample_full_rank(n_samples, self.mu, self.tril_elements)

    def get_hpd_interval(self):
        return NotImplementedError

class DoubleGaussianBNNPosteriorCPU(BaseGaussianBNNPosteriorCPU):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLL.__init__` docstring for the parameter description.

    """
    def __init__(self, Y_dim, Y_mean=None, Y_std=None):
        super(DoubleGaussianBNNPosteriorCPU, self).__init__(Y_dim, Y_mean, Y_std)
        self.tril_idx = np.tril_indices(self.Y_dim) # lower-triangular indices
        self.tril_len = len(self.tril_idx[0])
        self.out_dim = self.Y_dim**2 + 3*self.Y_dim + 1

    def set_sliced_pred(self, pred):
        d = self.Y_dim # for readability
        #pred = pred.cpu().numpy()
        self.batch_size = pred.shape[0]
        # First gaussian
        self.mu = pred[:, :d]
        self.tril_elements = pred[:, d:d+self.tril_len]
        self.mu2 = pred[:, d+self.tril_len:2*d+self.tril_len]
        self.tril_elements2 = pred[:, 2*d+self.tril_len:-1]
        self.w2 = 0.5*self.sigmoid(pred[:, -1].reshape(-1, 1))
        
    def sample(self, n_samples, sample_seed):
        """Sample from a mixture of two Gaussians, each with a full but constrained as low-rank plus diagonal covariance

        Parameters
        ----------
        n_samples : int
            how many samples to obtain
        sample_seed : int
            seed for the samples. Default: None

        Returns
        -------
        np.array of shape `[self.batch_size, n_samples, self.Y_dim]`
            samples

        """
        self.seed_samples(sample_seed)
        samples = np.zeros([self.batch_size, n_samples, self.Y_dim])
        # Determine first vs. second Gaussian
        unif2 = np.random.rand(self.batch_size, n_samples)
        second_gaussian = (self.w2 > unif2)
        # Sample from second Gaussian
        samples2 = self.sample_full_rank(n_samples, self.mu2, self.tril_elements2)
        samples[second_gaussian, :] = samples2[second_gaussian, :]
        # Sample from first Gaussian
        samples1 = self.sample_full_rank(n_samples, self.mu, self.tril_elements)
        samples[~second_gaussian, :] = samples1[~second_gaussian, :]
        return samples

    def get_hpd_interval(self):
        return NotImplementedError





import os
import shutil
import unittest
import numpy as np
import torch
import pandas as pd
from addict import Dict
from torch.utils.data import DataLoader
from h0rton.trainval_data import XYData
import h0rton.trainval_data.data_utils as data_utils

class TestDataUtils(unittest.TestCase):
    """"A suite of tests for data utility functions
    
    """
    @classmethod
    def setUpClass(cls):
        cls.Y_cols = ["lens_mass_center_x", "src_light_center_x","lens_mass_center_y", "src_light_center_y", "external_shear_gamma_ext", "external_shear_psi_ext"]
        cls.train_Y_mean = np.random.randn(len(cls.Y_cols))
        cls.train_Y_std = np.abs(np.random.randn(len(cls.Y_cols))) + 1.0
        cls.img_numpy = np.abs(np.random.randn(50*50)*2.0).reshape([1, 50, 50])
        cls.img_torch = torch.from_numpy(cls.img_numpy)
        n_data = 2
        cls.metadata = pd.DataFrame.from_dict({"lens_mass_center_x": np.random.randn(n_data), 
                                              "lens_mass_center_y": np.random.randn(n_data), 
                                              "src_light_center_x": np.random.randn(n_data), 
                                              "src_light_center_y": np.random.randn(n_data), 
                                              "external_shear_gamma_ext": np.random.randn(n_data),
                                                "external_shear_psi_ext": np.random.randn(n_data)
                                                })

    def test_whiten_pixels(self):
        """Test the torch pixel whitening vs. numpy

        """
        actual = data_utils.whiten_pixels(self.img_torch)
        expected = (self.img_numpy - np.mean(self.img_numpy))/np.std(self.img_numpy, ddof=1)
        np.testing.assert_array_almost_equal(actual, expected, err_msg='test_whiten_pixels')

    def test_asinh(self):
        """Test the torch asinh approximation vs. numpy

        """
        actual = data_utils.asinh(self.img_torch)
        expected = np.arcsinh(self.img_numpy)
        np.testing.assert_array_almost_equal(actual, expected, err_msg='test_asinh')

    def test_plus_1_log(self):
        """Test the torch log(1+X) vs. numpy

        """
        actual = data_utils.plus_1_log(self.img_torch)
        expected = np.log1p(self.img_numpy)
        np.testing.assert_array_almost_equal(actual, expected, err_msg='test_plus_1_log')

    def test_rescale_01(self):
        """Test the torch minmax stretching vs. numpy

        """
        actual = data_utils.rescale_01(self.img_torch)
        expected = (self.img_numpy - self.img_numpy.min())/(self.img_numpy.max() - self.img_numpy.min())
        np.testing.assert_array_almost_equal(actual, expected, err_msg='test_rescale_01')

    def test_whiten_Y_cols(self):
        """Test the Y whitening in pandas vs. numpy

        """
        # All columns
        actual = self.metadata.copy()
        data_utils.whiten_Y_cols(actual, self.train_Y_mean, self.train_Y_std, self.Y_cols)
        expected = (self.metadata[self.Y_cols].values - self.train_Y_mean.reshape([1, -1]))/self.train_Y_std.reshape([1, -1])
        np.testing.assert_array_almost_equal(actual[self.Y_cols].values, expected, err_msg='test_whiten_Y_cols')
        # Subset of columns
        actual = self.metadata.copy()
        subset_train_Y_mean = self.train_Y_mean[:3]
        subset_train_Y_std = self.train_Y_std[:3]
        subset_Y_cols = self.Y_cols[:3]
        data_utils.whiten_Y_cols(actual, subset_train_Y_mean, subset_train_Y_std, subset_Y_cols)
        expected = (self.metadata[subset_Y_cols].values - subset_train_Y_mean.reshape([1, -1]))/subset_train_Y_std.reshape([1, -1])
        np.testing.assert_array_almost_equal(actual[subset_Y_cols].values, expected, err_msg='test_whiten_Y_cols with a subset of the columns')

if __name__ == '__main__':
    unittest.main()


"""Script to combine the lenses for a joint H0 inference

Example
-------
To run this script, pass in the integer ID of the folder in which the individual D_dt samples are stored,
with the ID of the precision ceiling folder as the optional argument::
    
    $ python h0rton/combine_lenses.py 2 --prec_version_id 0

"""

import os
import numpy as np
import argparse
import pandas as pd
import lenstronomy
print(lenstronomy.__path__)
from h0rton.configs import TestConfig
import h0rton.h0_inference.h0_utils as h0_utils
from baobab.configs import BaobabConfig
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
import lenstronomy.Util.data_util as data_util
from baobab.sim_utils import Imager
from baobab.sim_utils import flux_utils, metadata_utils

def parse_args():
    """Parse command-line arguments
    
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('version_id', type=int, help='ID of the version folder in experiments')
    parser.add_argument('--prec_version_id', default=0, dest='prec_version_id', type=int,
                        help='ID of the version folder in experiments corresponding to the precision ceiling. Default: 0')
    parser.add_argument('--n_test', default=200, dest='n_test', type=int,
                        help='number of lenses to combine. Default: 200')
    args = parser.parse_args()
    return args

def main():
    args = parse_args()
    version_id = args.version_id
    prec_version_id = args.prec_version_id
    is_prec_ceiling = bool(version_id == prec_version_id)
    n_test = args.n_test

    version_dir = '/home/jwp/stage/sl/h0rton/experiments/v{:d}'.format(version_id)
    if is_prec_ceiling:
        test_cfg_path = os.path.join(version_dir, 'simple_mc_default.json')
    else:
        test_cfg_path = os.path.join(version_dir, 'mcmc_default.json')
    test_cfg = TestConfig.from_file(test_cfg_path)
    baobab_cfg = BaobabConfig.from_file(test_cfg.data.test_baobab_cfg_path)
    #train_val_cfg = TrainValConfig.from_file(test_cfg.train_val_config_file_path)
    # Read in truth metadata
    metadata = pd.read_csv(os.path.join(baobab_cfg.out_dir, 'metadata.csv'), index_col=None, nrows=n_test)
    # Read in summary
    summary = pd.read_csv(os.path.join(version_dir, 'summary.csv'), index_col=None, nrows=n_test)
    summary['id'] = summary.index
    true_Om0 = 0.3

    # Drop irrelevant lenses
    summary.drop(summary[summary['id']>(n_test - 1)].index, inplace=True)
    outside_rung = summary[summary['inference_time'] == 0].index
    summary.drop(outside_rung, inplace=True)
    print("Number of lenses being combined: {:d}".format(summary.shape[0]))
    print("Lenses that were discarded: ", set(np.arange(n_test)) - set(summary['id'].values))

    # Assign doubles vs quads
    summary['is_quad'] = (summary['n_img'] == 4)
    print("Doubles: ", len(summary[~summary['is_quad']]))
    print("Quads: ", len(summary[summary['is_quad']]))
    min_doubles_quads = np.min([len(summary[~summary['is_quad']]), len(summary[summary['is_quad']])])
    doubles = summary[~summary['is_quad']].iloc[:min_doubles_quads]
    quads = summary[summary['is_quad']].iloc[:min_doubles_quads]

    ####################
    # Lens combination #
    ####################
    if False:
        # 1. Gaussian D_dt
        print("Gaussian D_dt parameterization:")
        # Combine all lenses
        print("Combining all lenses...")
        combined_path_all = os.path.join(version_dir, 'combined_H0_summary.npy')
        _ = h0_utils.combine_lenses('DdtGaussian', ddt_mean=summary['D_dt_mean'].values, ddt_sigma=summary['D_dt_std'].values, z_lens=summary['z_lens'].values, z_src=summary['z_src'].values, samples_save_path=combined_path_all, corner_save_path=None, n_run=100, n_burn=500, n_walkers=20, true_Om0=true_Om0)
        # Combine all doubles
        print("Combining the {:d} doubles...".format(min_doubles_quads))
        combined_path_doubles = os.path.join(version_dir, 'combined_H0_doubles.npy')
        _ = h0_utils.combine_lenses('DdtGaussian', ddt_mean=doubles['D_dt_mean'].values, ddt_sigma=doubles['D_dt_std'].values, z_lens=doubles['z_lens'].values, z_src=doubles['z_src'].values, samples_save_path=combined_path_doubles, corner_save_path=None, n_run=100, n_burn=500, n_walkers=20, true_Om0=true_Om0)
        # Combine all quads
        print("Combining the {:d} quads...".format(min_doubles_quads))
        combined_path_quads = os.path.join(version_dir, 'combined_H0_quads.npy')
        _ = h0_utils.combine_lenses('DdtGaussian', ddt_mean=quads['D_dt_mean'].values, ddt_sigma=quads['D_dt_std'].values, z_lens=quads['z_lens'].values, z_src=quads['z_src'].values, samples_save_path=combined_path_quads, corner_save_path=None, n_run=100, n_burn=500, n_walkers=20, true_Om0=true_Om0)
        # 2. Lognormal D_dt
        print("Lognormal D_dt parameterization:")
        # Combine all lenses
        print("Combining all lenses...")
        combined_path_all_lognormal = os.path.join(version_dir, 'combined_H0_summary_lognormal.npy')
        _ = h0_utils.combine_lenses('DdtLogNorm', ddt_mu=summary['D_dt_mu'].values, ddt_sigma=summary['D_dt_sigma'].values, z_lens=summary['z_lens'].values, z_src=summary['z_src'].values, samples_save_path=combined_path_all_lognormal, corner_save_path=None, n_run=100, n_burn=500, n_walkers=20, true_Om0=true_Om0)
        # Combine all doubles
        print("Combining the {:d} doubles...".format(min_doubles_quads))
        combined_path_doubles_lognormal = os.path.join(version_dir, 'combined_H0_doubles_lognormal.npy')
        _ = h0_utils.combine_lenses('DdtLogNorm', ddt_mu=doubles['D_dt_mu'].values, ddt_sigma=doubles['D_dt_sigma'].values, z_lens=doubles['z_lens'].values, z_src=doubles['z_src'].values, samples_save_path=combined_path_doubles_lognormal, corner_save_path=None, n_run=100, n_burn=500, n_walkers=20, true_Om0=true_Om0)
        # Combine all quads
        print("Combining the {:d} quads...".format(min_doubles_quads))
        combined_path_quads_lognormal = os.path.join(version_dir, 'combined_H0_quads_lognormal.npy')
        _ = h0_utils.combine_lenses('DdtLogNorm', ddt_mu=quads['D_dt_mu'].values, ddt_sigma=quads['D_dt_sigma'].values, z_lens=quads['z_lens'].values, z_src=quads['z_src'].values, samples_save_path=combined_path_quads_lognormal, corner_save_path=None, n_run=100, n_burn=500, n_walkers=20, true_Om0=true_Om0)
        # 2. KDE
        print("KDE D_dt:")
        # Combine all lenses
        print("Combining all lenses...")
        combined_path_all_kde = os.path.join(version_dir, 'combined_H0_summary_kde.npy')
        _ = h0_utils.combine_lenses('DdtHistKDE', lens_ids=summary['id'].values, samples_dir=os.path.join(version_dir, 'mcmc_default'), z_lens=summary['z_lens'].values, z_src=summary['z_src'].values, samples_save_path=combined_path_all_kde, corner_save_path=None, n_run=100, n_burn=200, n_walkers=20, true_Om0=true_Om0, binning_method='scott')
        # Combine all doubles
        print("Combining the {:d} doubles...".format(min_doubles_quads))
        combined_path_doubles_kde = os.path.join(version_dir, 'combined_H0_doubles_kde.npy')
        _ = h0_utils.combine_lenses('DdtHistKDE', lens_ids=doubles['id'].values, samples_dir=os.path.join(version_dir, 'mcmc_default'), z_lens=doubles['z_lens'].values, z_src=doubles['z_src'].values, samples_save_path=combined_path_doubles_kde, corner_save_path=None, n_run=100, n_burn=500, n_walkers=20, true_Om0=true_Om0, binning_method='scott')
        # Combine all quads
        print("Combining the {:d} quads...".format(min_doubles_quads))
        combined_path_quads_kde = os.path.join(version_dir, 'combined_H0_quads_kde.npy')
        _ = h0_utils.combine_lenses('DdtHistKDE', lens_ids=quads['id'].values, samples_dir=os.path.join(version_dir, 'mcmc_default'), z_lens=quads['z_lens'].values, z_src=quads['z_src'].values, samples_save_path=combined_path_quads_kde, corner_save_path=None, n_run=100, n_burn=500, n_walkers=20, true_Om0=true_Om0, binning_method='scott')

    # Combine for each Einstein brightness bin
    if is_prec_ceiling:
        # If this version is the precision ceiling, compute the Einstein ring brightness for the first time.
        print("Computing Einstein ring brightness...")
        summary['lensed_E_ring_flux'] = 0.0
        summary['lensed_E_ring_mag'] = 0.0
        lens_mass_model = LensModel(lens_model_list=['PEMD', 'SHEAR_GAMMA_PSI'])
        src_light_model = LightModel(light_model_list=['SERSIC_ELLIPSE'])
        lens_light_model = LightModel(light_model_list=['SERSIC_ELLIPSE'])
        ps_model = PointSource(point_source_type_list=['LENSED_POSITION'], fixed_magnification_list=[False])
        components = ['lens_mass', 'src_light', 'agn_light', 'lens_light'] 
        for lens_i in range(n_test):
            imager = Imager(components, lens_mass_model, src_light_model, lens_light_model=lens_light_model, ps_model=ps_model, kwargs_numerics={'supersampling_factor': 1}, min_magnification=0.0, for_cosmography=True)
            imager._set_sim_api(num_pix=64, kwargs_detector=baobab_cfg.survey_object_dict[baobab_cfg.survey_info.bandpass_list[0]].kwargs_single_band(), psf_kernel_size=99, which_psf_maps=[101]) # TODO: read from BaobabConfig rather than hardcode
            imager.kwargs_src_light = metadata_utils.get_kwargs_src_light(metadata.iloc[lens_i])
            imager.kwargs_src_light = flux_utils.mag_to_amp_extended(imager.kwargs_src_light, imager.src_light_model, imager.data_api)
            imager.kwargs_lens_mass = metadata_utils.get_kwargs_lens_mass(metadata.iloc[lens_i])
            sample_ps = metadata_utils.get_nested_ps(metadata.iloc[lens_i])
            imager.for_cosmography = False
            imager._load_agn_light_kwargs(sample_ps)
            lensed_total_flux, lensed_src_img = flux_utils.get_lensed_total_flux(imager.kwargs_lens_mass, imager.kwargs_src_light, None, imager.image_model, return_image=True)
            lensed_ring_total_flux = np.sum(lensed_src_img)
            summary.loc[lens_i, 'lensed_E_ring_flux'] = lensed_ring_total_flux
            summary.loc[lens_i, 'lensed_E_ring_mag'] = data_util.cps2magnitude(lensed_ring_total_flux, baobab_cfg.survey_object_dict[baobab_cfg.survey_info.bandpass_list[0]].kwargs_single_band()['magnitude_zero_point'])
        lensed_ring_bins = np.quantile(summary['lensed_E_ring_mag'].values, [0.25, 0.5, 0.75, 1])
        lensed_ring_bins[-1] += 0.1 # buffer 
        summary['lensed_ring_bin'] = np.digitize(summary['lensed_E_ring_mag'].values, lensed_ring_bins)
        summary.to_csv(os.path.join(version_dir, 'ering_summary.csv'), index=False)
    else:
        # Simply read in the preciously computed Einstein ring brightness
        prec_version_dir = '/home/jwp/stage/sl/h0rton/experiments/v{:d}'.format(prec_version_id)
        prec_summary = pd.read_csv(os.path.join(prec_version_dir, 'ering_summary.csv'), index_col=None, nrows=n_test)
        summary['lensed_E_ring_mag'] = prec_summary['lensed_E_ring_mag'].values
        lensed_ring_bins = np.quantile(summary['lensed_E_ring_mag'].values, [0.25, 0.5, 0.75, 1])
        lensed_ring_bins[-1] += 0.1 # buffer 
        summary['lensed_ring_bin'] = np.digitize(summary['lensed_E_ring_mag'].values, lensed_ring_bins)
    
    for bin_i in range(len(lensed_ring_bins)):
        take_bin_i = (summary['lensed_ring_bin'] == bin_i)
        print("Combining {:d} lenses in bin {:d}...".format(len(summary[take_bin_i]), bin_i))
        if False:
            if is_prec_ceiling:
                combined_path_ering_i = os.path.join(version_dir, 'combined_H0_ering_{:d}.npy'.format(bin_i))
                _ = h0_utils.combine_lenses(likelihood_type='DdtGaussian', 
                                            ddt_mean=summary['D_dt_mean'][take_bin_i].values, 
                                            ddt_sigma=summary['D_dt_std'][take_bin_i].values, 
                                            z_lens=summary['z_lens'][take_bin_i].values, 
                                            z_src=summary['z_src'][take_bin_i].values, 
                                            samples_save_path=combined_path_ering_i, 
                                            corner_save_path=None, 
                                            n_run=100, 
                                            n_burn=500, 
                                            n_walkers=20, 
                                            true_Om0=true_Om0)
            combined_path_ering_i_lognormal = os.path.join(version_dir, 'combined_H0_ering_{:d}_lognormal.npy'.format(bin_i))
            _ = h0_utils.combine_lenses(likelihood_type='DdtLogNorm', 
                                        ddt_mu=summary['D_dt_mu'][take_bin_i].values, 
                                        ddt_sigma=summary['D_dt_sigma'][take_bin_i].values, 
                                        z_lens=summary['z_lens'][take_bin_i].values, 
                                        z_src=summary['z_src'][take_bin_i].values, 
                                        samples_save_path=combined_path_ering_i_lognormal, 
                                        corner_save_path=None, 
                                        n_run=100, 
                                        n_burn=500, 
                                        n_walkers=20, 
                                        true_Om0=true_Om0)
        combined_path_ering_i_kde = os.path.join(version_dir, 'combined_H0_ering_{:d}_kde.npy'.format(bin_i))
        _ = h0_utils.combine_lenses(likelihood_type='DdtHistKDE', 
                                    lens_ids=summary['id'][take_bin_i].values, 
                                    samples_dir=os.path.join(version_dir, 'mcmc_default'),
                                    z_lens=summary['z_lens'][take_bin_i].values, 
                                    z_src=summary['z_src'][take_bin_i].values,  
                                    samples_save_path=combined_path_ering_i_kde, 
                                    corner_save_path=None, 
                                    n_run=100, 
                                    n_burn=500, 
                                    n_walkers=20, 
                                    true_Om0=true_Om0, 
                                    binning_method='scott')

if __name__ == '__main__':
    main()

    





# -*- coding: utf-8 -*-
"""Training the Bayesian neural network (BNN).
This script trains the BNN according to the config specifications.

Example
-------
To run this script, pass in the path to the user-defined training config file as the argument::
    
    $ train h0rton/example_user_config.py

"""

import os, sys
import random
import argparse
from addict import Dict
import numpy as np # linear algebra
from tqdm import tqdm
# torch modules
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
# h0rton modules
from h0rton.trainval_data import XYData, XYTestData
from h0rton.configs import TrainValConfig
import h0rton.losses
import h0rton.models
import h0rton.h0_inference
import h0rton.train_utils as train_utils

def parse_args():
    """Parse command-line arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('user_cfg_path', help='path to the user-defined training config file')
    #parser.add_argument('--n_data', default=None, dest='n_data', type=int,
    #                    help='size of dataset to generate (overrides config file)')
    args = parser.parse_args()
    # sys.argv rerouting for setuptools entry point
    if args is None:
        args = Dict()
        args.user_cfg_path = sys.argv[0]
        #args.n_data = sys.argv[1]
    return args

def seed_everything(global_seed):
    """Seed everything for reproducibility

    global_seed : int
        seed for `np.random`, `random`, and relevant `torch` backends

    """
    np.random.seed(global_seed)
    random.seed(global_seed)
    torch.manual_seed(global_seed)
    torch.cuda.manual_seed_all(global_seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False

def main():
    args = parse_args()
    cfg = TrainValConfig.from_file(args.user_cfg_path)
    # Set device and default data type
    device = torch.device(cfg.device_type)
    if device.type == 'cuda':
        torch.set_default_tensor_type('torch.cuda.' + cfg.data.float_type)
    else:
        torch.set_default_tensor_type('torch.' + cfg.data.float_type)
    train_utils.seed_everything(cfg.global_seed)

    ############
    # Data I/O #
    ############

    # Define training data and loader
    torch.multiprocessing.set_start_method('spawn', force=True)
    train_data = XYData(cfg.data.train_dir, data_cfg=cfg.data)
    train_loader = DataLoader(train_data, batch_size=cfg.optim.batch_size, shuffle=True, drop_last=True, num_workers=6, pin_memory=True)
    n_train = train_data.n_data - (train_data.n_data % cfg.optim.batch_size)

    # Define val data and loader
    val_data = XYData(cfg.data.val_dir, data_cfg=cfg.data)
    val_loader = DataLoader(val_data, batch_size=min(len(val_data), cfg.optim.batch_size), shuffle=True, drop_last=True,)
    n_val = val_data.n_data - (val_data.n_data % min(len(val_data), cfg.optim.batch_size))

    #########
    # Model #
    #########
    Y_dim = cfg.data.Y_dim
    # Instantiate loss function
    loss_fn = getattr(h0rton.losses, cfg.model.likelihood_class)(Y_dim=Y_dim, device=device)
    # Instantiate posterior (for logging)
    #bnn_post = getattr(h0rton.h0_inference.gaussian_bnn_posterior, loss_fn.posterior_name)(val_data.Y_dim, device, val_data.train_Y_mean, val_data.train_Y_std)
    # Instantiate model
    net = getattr(h0rton.models, cfg.model.architecture)(num_classes=loss_fn.out_dim, dropout_rate=cfg.model.dropout_rate, norm_layer=h0rton.models.BatchNorm2d)
    net.to(device)

    ################
    # Optimization #
    ################

    # Instantiate optimizer
    optimizer = optim.Adam(net.parameters(), lr=cfg.optim.learning_rate, amsgrad=False, weight_decay=cfg.optim.weight_decay)
    lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=cfg.optim.lr_scheduler.factor, patience=cfg.optim.lr_scheduler.patience, verbose=True)
    
    # Saving/loading state dicts
    checkpoint_dir = cfg.checkpoint.save_dir
    if not os.path.exists(checkpoint_dir):
        os.mkdir(checkpoint_dir)

    if cfg.model.load_state:
        epoch, net, optimizer, train_loss, val_loss = train_utils.load_state_dict(cfg.model.state_path, net, optimizer, cfg.optim.n_epochs, device)
        epoch += 1 # resume with next epoch
        last_saved_val_loss = val_loss
        #print(lr_scheduler.state_dict())
        #print(optimizer.state_dict())
    else:
        epoch = 0
        last_saved_val_loss = np.inf

    logger = SummaryWriter()
    model_path = ''
    print("Training set size: {:d}".format(n_train))
    print("Validation set size: {:d}".format(n_val))
    if cfg.data.test_dir is not None:
        print("Test set size: {:d}".format(n_test))
    
    progress = tqdm(range(epoch, cfg.optim.n_epochs))
    for epoch in progress:
        net.train()
        #net.apply(h0rton.models.deactivate_batchnorm)
        train_loss = 0.0

        for batch_idx, (X_tr, Y_tr) in enumerate(train_loader):
            X_tr = X_tr.to(device)
            Y_tr = Y_tr.to(device)
            # Update weights
            optimizer.zero_grad()
            pred_tr = net.forward(X_tr)
            loss = loss_fn(pred_tr, Y_tr)
            loss.backward()
            optimizer.step()
            # For logging
            train_loss += (loss.detach().item() - train_loss)/(1 + batch_idx)
        tqdm.write("Epoch [{}/{}]: TRAIN Loss: {:.4f}".format(epoch+1, cfg.optim.n_epochs, train_loss))

        net.eval()         
        with torch.no_grad():
            #net.apply(h0rton.models.deactivate_batchnorm)
            val_loss = 0.0
            test_loss = 0.0

            if cfg.data.test_dir is not None:
                for batch_idx, (X_t, Y_t) in enumerate(test_loader):
                    X_t = X_t.to(device)
                    Y_t = Y_t.to(device)
                    pred_t = net.forward(X_t)
                    nograd_loss_t = loss_fn(pred_t, Y_t)
                    test_loss += (nograd_loss_t.detach().item() - test_loss)/(1 + batch_idx)
   
            for batch_idx, (X_v, Y_v) in enumerate(val_loader):
                X_v = X_v.to(device)
                Y_v = Y_v.to(device)
                pred_v = net.forward(X_v)
                nograd_loss_v = loss_fn(pred_v, Y_v)
                val_loss += (nograd_loss_v.detach().item() - val_loss)/(1 + batch_idx)

            tqdm.write("Epoch [{}/{}]: VALID Loss: {:.4f}".format(epoch+1, cfg.optim.n_epochs, val_loss))
            if cfg.data.test_dir is not None:
                tqdm.write("Epoch [{}/{}]: TEST Loss: {:.4f}".format(epoch+1, cfg.optim.n_epochs, test_loss))
            
            if (epoch + 1)%(cfg.monitoring.interval) == 0:
                # Log train and val metrics
                loss_dict = {'train': train_loss, 'val': val_loss}
                if cfg.data.test_dir is not None:
                    loss_dict.update(test=test_loss)
                logger.add_scalars('metrics/loss', loss_dict, epoch)

            if (epoch + 1)%(cfg.checkpoint.interval) == 0:
                # FIXME compare to last saved epoch val loss
                if val_loss < last_saved_val_loss:
                    os.remove(model_path) if os.path.exists(model_path) else None
                    model_path = train_utils.save_state_dict(net, optimizer, lr_scheduler, train_loss, val_loss, checkpoint_dir, cfg.model.architecture, epoch)
                    last_saved_val_loss = val_loss
        lr_scheduler.step(train_loss)

    logger.close()
    # Save final state dict
    if val_loss < last_saved_val_loss:
        os.remove(model_path) if os.path.exists(model_path) else None
        model_path = train_utils.save_state_dict(net, optimizer, lr_scheduler, train_loss, val_loss, checkpoint_dir, cfg.model.architecture, epoch)
        print("Saved model at {:s}".format(os.path.abspath(model_path)))

if __name__ == '__main__':
    main()


from abc import ABC, abstractmethod
import numpy as np
from scipy.special import logsumexp
from h0rton.h0_inference.gaussian_bnn_posterior_cpu import sigmoid, logsigmoid
__all__ = ['BaseGaussianNLLCPU', 'DiagonalGaussianNLLCPU', 'FullRankGaussianNLLCPU', 'DoubleGaussianNLLCPU']

log_2_pi = 1.8378770664093453
log_2 = 0.6931471805599453

class BaseGaussianNLLCPU(ABC):
    """Abstract base class to represent the Gaussian negative log likelihood (NLL).

    Gaussian NLLs or mixtures thereof with various forms of the covariance matrix inherit from this class.

    """
    def __init__(self, Y_dim):
        """
        Parameters
        ----------
        Y_dim : int
            number of parameters to predict

        """
        self.Y_dim = Y_dim
        self.sigmoid = sigmoid
        self.logsigmoid = logsigmoid

    @abstractmethod
    def slice(self, pred):
        """Slice the raw network prediction into meaningful Gaussian parameters

        Parameters
        ----------
        pred : np.Tensor of shape `[batch_size, self.Y_dim]`
            the network prediction

        """
        return NotImplemented

    @abstractmethod
    def __call__(self, pred, target):
        """Evaluate the NLL. Must be overridden by subclasses.

        Parameters
        ----------
        pred : np.Tensor
            raw network output for the predictions
        target : np.Tensor
            Y labels

        """
        return NotImplemented

    def nll_diagonal(self, target, mu, logvar):
        """Evaluate the NLL for single Gaussian with diagonal covariance matrix

        Parameters
        ----------
        target : np.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : np.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior
        logvar : np.Tensor of shape [batch_size, Y_dim]
            network prediction of the log of the diagonal elements of the covariance matrix

        Returns
        -------
        np.Tensor of shape
            NLL values

        """
        precision = np.exp(-logvar)
        # Loss kernel
        loss = precision * (target - mu)**2.0 + logvar
        # Restore prefactors
        loss += np.log(2.0*np.pi)
        loss *= 0.5
        return np.mean(np.sum(loss, axis=1), axis=0)

    def nll_full_rank(self, target, mu, tril_elements, reduce=True):
        """Evaluate the NLL for a single Gaussian with a full-rank covariance matrix

        Parameters
        ----------
        target : np.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : np.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior
        tril_elements : np.Tensor of shape [batch_size, Y_dim*(Y_dim + 1)//2]
        reduce : bool
            whether to take the mean across the batch

        Returns
        -------
        np.Tensor of shape [batch_size,]
            NLL values

        """
        batch_size, _ = target.shape
        tril = np.zeros([batch_size, self.Y_dim, self.Y_dim])
        tril[:, self.tril_idx[0], self.tril_idx[1]] = tril_elements
        log_diag_tril = np.diagonal(tril, offset=0, axis1=1, axis2=2) # [batch_size, Y_dim]
        logdet_term = -np.sum(log_diag_tril, axis=1) # [batch_size,]
        tril[:, np.eye(self.Y_dim).astype(bool)] = np.exp(log_diag_tril)
        prec_mat = np.matmul(tril, np.transpose(tril, [0, 2, 1])) # [batch_size, Y_dim, Y_dim]
        y_diff = mu - target # [batch_size, Y_dim]
        mahalanobis_term = 0.5*np.sum(
            y_diff*np.sum(prec_mat*np.expand_dims(y_diff, -1), axis=-2), axis=-1) # [batch_size,]
        loss = logdet_term + mahalanobis_term + 0.5*self.Y_dim*log_2_pi
        if reduce:
            return np.mean(loss, axis=0) # float
        else:
            return loss # [batch_size,]

    def nll_mixture(self, target, mu, tril_elements, mu2, tril_elements2, alpha):
        """Evaluate the NLL for a single Gaussian with a full but low-rank plus diagonal covariance matrix

        Parameters
        ----------
        target : np.Tensor of shape [batch_size, Y_dim]
            Y labels
        mu : np.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior for the first Gaussian
        tril_elements : np.Tensor of shape [batch_size, self.tril_len]
            network prediction of the elements in the precision matrix
        mu2 : np.Tensor of shape [batch_size, Y_dim]
            network prediction of the mu (mean parameter) of the BNN posterior for the second Gaussian
        tril_elements2 : np.Tensor of shape [batch_size, self.tril_len]
            network prediction of the elements in the precision matrix for the second Gaussian
        alpha : np.Tensor of shape [batch_size, 1]
            network prediction of the logit of twice the weight on the second Gaussian 

        Note
        ----
        The weight on the second Gaussian is required to be less than 0.5, to make the two Gaussians well-defined.

        Returns
        -------
        np.Tensor of shape [batch_size,]
            NLL values

        """
        batch_size, _ = target.shape
        log_ll = np.empty([batch_size, 2], dtype=None)
        alpha = alpha.reshape(-1)
        log_ll[:, 0] = np.log1p(2.0*np.exp(-alpha)) - log_2 - np.log1p(np.exp(-alpha)) - self.nll_full_rank(target, mu, tril_elements, reduce=False) # [batch_size]
        # np.log(np.tensor([0.5])).double()
        log_ll[:, 1] = -log_2 + self.logsigmoid(alpha) - self.nll_full_rank(target, mu2, tril_elements2, reduce=False) # [batch_size], 0.6931471 = np.log(2)
        log_nll = -logsumexp(log_ll, axis=1)
        return np.mean(log_nll)

class DiagonalGaussianNLLCPU(BaseGaussianNLLCPU):
    """The negative log likelihood (NLL) for a single Gaussian with diagonal covariance matrix
        
    `BaseGaussianNLLCPU.__init__` docstring for the parameter description.

    """
    posterior_name = 'DiagonalGaussianBNNPosterior'

    def __init__(self, Y_dim):
        super(DiagonalGaussianNLLCPU, self).__init__(Y_dim)
        self.out_dim = Y_dim*2

    def __call__(self, pred, target):
        sliced = self.slice(pred)
        return self.nll_diagonal(target, **sliced)

    def slice(self, pred):
        d = self.Y_dim # for readability
        sliced = dict(
                      mu=pred[:, :d],
                      logvar=pred[:, d:]
                      )
        return sliced

class FullRankGaussianNLLCPU(BaseGaussianNLLCPU):
    """The negative log likelihood (NLL) for a single Gaussian with a full-rank covariance matrix
        
    See `BaseGaussianNLLCPU.__init__` docstring for the parameter description.

    """
    posterior_name = 'FullRankGaussianBNNPosterior'

    def __init__(self, Y_dim):
        super(FullRankGaussianNLLCPU, self).__init__(Y_dim)
        self.tril_idx = np.tril_indices(self.Y_dim) # lower-triangular indices
        self.tril_len = len(self.tril_idx[0])
        self.out_dim = self.Y_dim + self.Y_dim*(self.Y_dim + 1)//2

    def __call__(self, pred, target):
        sliced = self.slice(pred)
        return self.nll_full_rank(target, **sliced, reduce=True)

    def slice(self, pred):
        d = self.Y_dim # for readability
        sliced = dict(
                      mu=pred[:, :d],
                      tril_elements=pred[:, d:d+self.tril_len]
                      )
        return sliced

class DoubleGaussianNLLCPU(BaseGaussianNLLCPU):
    """The negative log likelihood (NLL) for a mixture of two Gaussians, each with a full but constrained as low-rank plus diagonal covariance 
        
    Only rank 2 is currently supported. `BaseGaussianNLLCPU.__init__` docstring for the parameter description.

    """
    posterior_name = 'DoubleGaussianBNNPosterior'

    def __init__(self, Y_dim):
        super(DoubleGaussianNLLCPU, self).__init__(Y_dim)
        self.tril_idx = np.tril_indices(self.Y_dim) # lower-triangular indices
        self.tril_len = len(self.tril_idx[0])
        self.out_dim = self.Y_dim**2 + 3*self.Y_dim + 1

    def __call__(self, pred, target):
        sliced = self.slice(pred)
        return self.nll_mixture(target, **sliced)

    def slice(self, pred):
        d = self.Y_dim # for readability
        sliced = dict(
                      mu=pred[:, :d],
                      tril_elements=pred[:, d:d+self.tril_len],
                      mu2=pred[:, d+self.tril_len:2*d+self.tril_len],
                      tril_elements2=pred[:, 2*d+self.tril_len:-1],
                      alpha=pred[:, -1]
                      )
        return sliced


# -*- coding: utf-8 -*-
"""Training the Bayesian neural network (BNN).
This script trains the BNN according to the config specifications.

Example
-------
To run this script, pass in the path to the user-defined training config file as the argument::
    
    $ train h0rton/example_user_config.py

"""

import os, sys
import random
import argparse
from addict import Dict
import numpy as np # linear algebra
from tqdm import tqdm
# torch modules
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
# h0rton modules
from h0rton.trainval_data import XYData
from h0rton.configs import TrainValConfig
import h0rton.losses
import h0rton.models
import h0rton.h0_inference
import h0rton.train_utils as train_utils
import h0rton.script_utils as script_utils

def parse_args():
    """Parse command-line arguments
    
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('user_cfg_path', help='path to the user-defined training config file')
    #parser.add_argument('--n_data', default=None, dest='n_data', type=int,
    #                    help='size of dataset to generate (overrides config file)')
    args = parser.parse_args()
    # sys.argv rerouting for setuptools entry point
    if args is None:
        args = Dict()
        args.user_cfg_path = sys.argv[0]
        #args.n_data = sys.argv[1]
    return args

def main():
    args = parse_args()
    cfg = TrainValConfig.from_file(args.user_cfg_path)
    # Set device and default data type
    device = torch.device(cfg.device_type)
    if device.type == 'cuda':
        torch.set_default_tensor_type('torch.cuda.' + cfg.data.float_type)
    else:
        torch.set_default_tensor_type('torch.' + cfg.data.float_type)
    script_utils.seed_everything(cfg.global_seed)

    ############
    # Data I/O #
    ############

    # Define training data and loader
    #torch.multiprocessing.set_start_method('spawn', force=True)
    train_data = XYData(is_train=True, 
                        Y_cols=cfg.data.Y_cols, 
                        float_type=cfg.data.float_type, 
                        define_src_pos_wrt_lens=cfg.data.define_src_pos_wrt_lens, 
                        rescale_pixels=cfg.data.rescale_pixels, 
                        log_pixels=cfg.data.log_pixels, 
                        add_pixel_noise=cfg.data.add_pixel_noise, 
                        eff_exposure_time=cfg.data.eff_exposure_time, 
                        train_Y_mean=None, 
                        train_Y_std=None, 
                        train_baobab_cfg_path=cfg.data.train_baobab_cfg_path, 
                        val_baobab_cfg_path=cfg.data.val_baobab_cfg_path, 
                        for_cosmology=False)
    train_loader = DataLoader(train_data, batch_size=cfg.optim.batch_size, shuffle=True, drop_last=True)
    n_train = len(train_data) - (len(train_data) % cfg.optim.batch_size)

    # Define val data and loader
    val_data = XYData(is_train=False, 
                      Y_cols=cfg.data.Y_cols, 
                      float_type=cfg.data.float_type, 
                      define_src_pos_wrt_lens=cfg.data.define_src_pos_wrt_lens, 
                      rescale_pixels=cfg.data.rescale_pixels, 
                      log_pixels=cfg.data.log_pixels, 
                      add_pixel_noise=cfg.data.add_pixel_noise, 
                      eff_exposure_time=cfg.data.eff_exposure_time, 
                      train_Y_mean=train_data.train_Y_mean, 
                      train_Y_std=train_data.train_Y_std, 
                      train_baobab_cfg_path=cfg.data.train_baobab_cfg_path, 
                      val_baobab_cfg_path=cfg.data.val_baobab_cfg_path, 
                      for_cosmology=False)
    val_loader = DataLoader(val_data, batch_size=min(len(val_data), cfg.optim.batch_size), shuffle=False, drop_last=True,)
    n_val = len(val_data) - (len(val_data) % min(len(val_data), cfg.optim.batch_size))

    #########
    # Model #
    #########
    Y_dim = val_data.Y_dim
    # Instantiate loss function
    loss_fn = getattr(h0rton.losses, cfg.model.likelihood_class)(Y_dim=Y_dim, device=device)
    # Instantiate posterior (for logging)
    bnn_post = getattr(h0rton.h0_inference.gaussian_bnn_posterior, loss_fn.posterior_name)(val_data.Y_dim, device, val_data.train_Y_mean, val_data.train_Y_std)
    # Instantiate model
    net = getattr(h0rton.models, cfg.model.architecture)(num_classes=loss_fn.out_dim, dropout_rate=cfg.model.dropout_rate)
    net.to(device)

    ################
    # Optimization #
    ################

    # Instantiate optimizer
    optimizer = optim.Adam(net.parameters(), lr=cfg.optim.learning_rate, amsgrad=False, weight_decay=cfg.optim.weight_decay)
    #optimizer = optim.SGD(net.parameters(), lr=cfg.optim.learning_rate, weight_decay=cfg.optim.weight_decay)
    lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.75, patience=50, cooldown=50, min_lr=1e-5, verbose=True)
    #lr_scheduler = optim.lr_scheduler.CyclicLR(optimizer, base_lr=cfg.optim.learning_rate*0.2, max_lr=cfg.optim.learning_rate, step_size_up=cfg.optim.lr_scheduler.step_size_up, step_size_down=None, mode='triangular2', gamma=1.0, scale_fn=None, scale_mode='cycle', cycle_momentum=True, base_momentum=0.8, max_momentum=0.9, last_epoch=-1)
    
    # Saving/loading state dicts
    checkpoint_dir = cfg.checkpoint.save_dir
    if not os.path.exists(checkpoint_dir):
        os.mkdir(checkpoint_dir)

    if cfg.model.load_state:
        epoch, net, optimizer, train_loss, val_loss = train_utils.load_state_dict(cfg.model.state_path, net, optimizer, cfg.optim.n_epochs, device)
        epoch += 1 # resume with next epoch
        last_saved_val_loss = val_loss
        print(lr_scheduler.state_dict())
        print(optimizer.state_dict())
    else:
        epoch = 0
        last_saved_val_loss = np.inf

    logger = SummaryWriter()
    model_path = ''
    print("Training set size: {:d}".format(n_train))
    print("Validation set size: {:d}".format(n_val))
    
    progress = tqdm(range(epoch, cfg.optim.n_epochs))
    n_iter = 0
    for epoch in progress:
        #net.apply(h0rton.models.deactivate_batchnorm)
        train_loss = 0.0
        for batch_idx, (X_tr, Y_tr) in enumerate(train_loader):
            n_iter += 1
            net.train()
            X_tr = X_tr.to(device)
            Y_tr = Y_tr.to(device)
            # Update weights
            optimizer.zero_grad()
            pred_tr = net.forward(X_tr)
            loss = loss_fn(pred_tr, Y_tr)
            loss.backward()
            optimizer.step()
            # For logging
            train_loss += (loss.detach().item() - train_loss)/(1 + batch_idx)
            # Step lr_scheduler every batch
            lr_scheduler.step(train_loss)
            tqdm.write("Iter [{}/{}/{}]: TRAIN Loss: {:.4f}".format(n_iter, epoch+1, cfg.optim.n_epochs, train_loss))

            if (n_iter)%(cfg.monitoring.interval) == 0:
                net.eval()         
                with torch.no_grad():
                    #net.apply(h0rton.models.deactivate_batchnorm)
                    val_loss = 0.0
           
                    for batch_idx, (X_v, Y_v) in enumerate(val_loader):
                        X_v = X_v.to(device)
                        Y_v = Y_v.to(device)
                        pred_v = net.forward(X_v)
                        nograd_loss_v = loss_fn(pred_v, Y_v)
                        val_loss += (nograd_loss_v.detach().item() - val_loss)/(1 + batch_idx)

                    tqdm.write("Epoch [{}/{}]: VALID Loss: {:.4f}".format(epoch+1, cfg.optim.n_epochs, val_loss))
                    
                    # Subset of validation for plotting
                    n_plotting = cfg.monitoring.n_plotting
                    #X_plt = X_v[:n_plotting].cpu().numpy()
                    #Y_plt = Y[:n_plotting].cpu().numpy()
                    Y_plt_orig = bnn_post.transform_back_mu(Y_v[:n_plotting]).cpu().numpy()
                    pred_plt = pred_v[:n_plotting]
                    # Slice pred_plt into meaningful Gaussian parameters for this batch
                    bnn_post.set_sliced_pred(pred_plt)
                    mu_orig = bnn_post.transform_back_mu(bnn_post.mu).cpu().numpy()
                    # Log train and val metrics
                    loss_dict = {'train': train_loss, 'val': val_loss}
                    logger.add_scalars('metrics/loss', loss_dict, n_iter)
                    #mae = train_utils.get_mae(mu, Y_plt)
                    mae_dict = train_utils.get_mae(mu_orig, Y_plt_orig, cfg.data.Y_cols)
                    logger.add_scalars('metrics/mae', mae_dict, n_iter)
                    # Log log determinant of the covariance matrix
                    
                    if cfg.model.likelihood_class in ['DoubleGaussianNLL', 'FullRankGaussianNLL']:
                        logdet = train_utils.get_logdet(bnn_post.tril_elements.cpu().numpy(), Y_dim)
                        logger.add_histogram('logdet_cov_mat', logdet, n_iter)
                    # Log second Gaussian stats
                    if cfg.model.likelihood_class in ['DoubleGaussianNLL', 'DoubleLowRankGaussianNLL']:
                        # Log histogram of w2
                        logger.add_histogram('val_pred/weight_gaussian2', bnn_post.w2.cpu().numpy(), n_iter)
                        # Log RMSE of second Gaussian
                        mu2_orig = bnn_post.transform_back_mu(bnn_post.mu2).cpu().numpy()
                        mae2_dict = train_utils.get_mae(mu2_orig, Y_plt_orig, cfg.data.Y_cols)
                        logger.add_scalars('metrics/mae2', mae2_dict, n_iter)
                        # Log logdet of second Gaussian
                        logdet2 = train_utils.get_logdet(bnn_post.tril_elements2.cpu().numpy(), Y_dim)
                        logger.add_histogram('logdet_cov_mat2', logdet2, n_iter)

                    if val_loss < last_saved_val_loss:
                        os.remove(model_path) if os.path.exists(model_path) else None
                        model_path = train_utils.save_state_dict(net, optimizer, lr_scheduler, train_loss, val_loss, checkpoint_dir, cfg.model.architecture, epoch)
                        last_saved_val_loss = val_loss

    logger.close()
    # Save final state dict
    if val_loss < last_saved_val_loss:
        os.remove(model_path) if os.path.exists(model_path) else None
        model_path = train_utils.save_state_dict(net, optimizer, lr_scheduler, train_loss, val_loss, checkpoint_dir, cfg.model.architecture, epoch)
        print("Saved model at {:s}".format(os.path.abspath(model_path)))

if __name__ == '__main__':
    main()


import os
import numpy as np
import pandas as pd
from ast import literal_eval
import re
import h0rton.tdlmc_data
from h0rton.tdlmc_utils import tdlmc_metrics

__all__ = ['convert_to_dataframe', 'parse_closed_box', 'parse_open_box', 
'read_from_csv', 'format_results_for_tdlmc_metrics']

tdlmc_data_path = os.path.abspath(list(h0rton.tdlmc_data.__path__)[0])
"""str: directory path containing the TDLMC data

"""

def format_results_for_tdlmc_metrics(version_dir, out_dir, rung_id=2):
    """Format the BNN inference results so they can be read into the script that 
    generates the TDLMC metrics cornerplot

    Parameters
    ----------
    version_dir : str or os.path object
        path to the folder containing inference results
    rung_id : int
        TDLMC rung ID

    """
    label_to_id = {'A1 (0.5 HST orbit)': (4, ''), 
                   'A2 (1 HST orbit)': (3, ''), 
                   'A3 (2 HST orbits)': (2, ''), 
                   'B1 (89 doubles for 1 HST orbit)': (3, '_doubles'), 
                   'B2 (89 quads for 1 HST orbit)': (3, '_quads')}

    for label, (version_id, img) in label_to_id.items():
        summary = pd.read_csv(os.path.join(version_dir, 'summary.csv'), 
                              index_col=None)
        true_H0 = 70.0
        outside_rung = summary[ summary['id'] > (199)].index
        summary.drop(list(outside_rung), inplace=True)
        summary['keep'] = True # keep all lenses
        summary.loc[~summary['keep'], ['H0_mean', 'H0_std']] = -99
        summary['id'] = summary['id'].astype(int)
        
        if img != '':
            summary['is_quad'] = (summary['n_img'] == 4)
            n_test = np.min([len(summary[~summary['is_quad']]), 
                            len(summary[summary['is_quad']])])
            if img == '_doubles':
                summary = summary[~summary['is_quad']].iloc[:n_test]
            else:
                summary = summary[summary['is_quad']].iloc[:n_test]
                
        tdlmc_mean = summary['H0_mean'][summary['keep']]
        tdlmc_std = summary['H0_std'][summary['keep']]

        # Compute per-lens versions of the metrics
        summary['g'] = ((summary['H0_mean'] - true_H0)/summary['H0_std'])**2.0
        summary['log_g'] = np.log10(summary['g'])
        summary['p'] = (summary['H0_std']/true_H0)
        summary['a'] = (summary['H0_mean'] - true_H0)/true_H0

        # Test-set-side metrics
        G = tdlmc_metrics.get_goodness(tdlmc_mean,tdlmc_std, true_H0)
        P = tdlmc_metrics.get_precision(tdlmc_std, true_H0)
        A = tdlmc_metrics.get_accuracy(tdlmc_mean, true_H0)
        print("Goodness: ", G, "Log goodness: ", np.log10(G))
        print("Precision: ", P)
        print("Accuracy: ", A)
        print("Total combined", summary[summary['keep']].shape[0])
        print("Actually discarded", summary[~summary['keep']].shape[0])
        lens_name_formatting = lambda x: 'rung{:d}_seed{:d}'.format(rung_id, x)
        summary['rung_id'] = summary.id.apply(lens_name_formatting)
        summary = summary[['rung_id', 'H0_mean', 'H0_std']]
        summary.to_csv(os.path.join(out_dir, 'H0rton/{:s}.txt'.format(label)), 
                       header=None, index=None, sep=' ', mode='a')

def read_from_csv(csv_path):
    """Read a Pandas Dataframe from the combined csv file of TDLMC data while 
    evaluating all the relevant strings in each column as Python objects

    Parameters
    ----------
    csv_path : str
        path to the csv file generated using `convert_to_dataframe`

    Returns
    -------
    Pandas DataFrame
        the TDLMC data with correct Python objects

    """
    df = pd.read_csv(csv_path, index_col=False)
    # These are columns that are lists
    for list_col in [
                    'host_pos', 
                    'measured_td', 
                    'measured_td_err',
                    'agn_img_pos_x', 
                    'agn_img_pos_y', 
                    'agn_img_amp', 
                    'time_delays',
                    ]:
        df[list_col] = df[list_col].apply(literal_eval).apply(np.array)
    return df

def convert_to_dataframe(rung, save_csv_path):
    """Store the TDLMC closed and open boxes into a Pandas DataFrame and exports 
    to a csv file at the same location

    Parameters
    ----------
    rung : int
        rung number
    save_csv_path : str
        path of the csv file to be generated

    Returns
    -------
    Pandas DataFrame
        the extracted rung data

    """
    if save_csv_path is None:
        save_csv_path = os.path.join(tdlmc_data_path, 'rung{:d}_combined.csv'.format(rung))
        print("Saving rung {:d} data at {:s}...".format(rung, save_csv_path))

    df = pd.DataFrame()
    for code in ['code1', 'code2']:
        closed_code_dir = os.path.join(tdlmc_data_path, 'rung{:d}'.format(rung), code)
        open_code_dir = os.path.join(tdlmc_data_path, 'rung{:d}_open_box'.format(rung), code)
        seeds = sorted(os.listdir(closed_code_dir)) # list of seeds, e.g. 'f160w-seed101'
        row = {} # initialized dict in which to save lens info
        for seed in seeds:
            # Path to the text files
            closed_box_path = os.path.join(closed_code_dir, seed, 'lens_info_for_Good_team.txt')
            open_box_path = os.path.join(open_code_dir, seed, 'lens_all_info.txt')
            # Save seed path for easy access
            row['name'] = 'rung{:d}_{:s}_{:s}'.format(rung, code, seed)
            row['seed'] = seed
            row['seed_path'] = os.path.join(closed_code_dir, seed)
            # Parse the text files
            row = parse_closed_box(closed_box_path, row)
            row = parse_open_box(open_box_path, row)
            df = df.append(row, ignore_index=True)
    # Unravel nested dictionaries in some columns
    lens_mass = df['lens_mass'].apply(pd.Series).copy().add_prefix('lens_mass_')
    lens_light = df['lens_light'].apply(pd.Series).copy().add_prefix('lens_light_')
    ext_shear_bphi = df['ext_shear_bphi'].apply(pd.Series).copy().add_prefix('ext_shear_')
    ext_shear_e1e2 =  df['ext_shear_e1e2'].apply(pd.Series).copy().add_prefix('ext_shear_')
    df = pd.concat([df.drop(['lens_mass', 'lens_light', 'ext_shear_bphi', 'ext_shear_e1e2'], axis=1), lens_mass, lens_light, ext_shear_bphi, ext_shear_e1e2], axis=1)

    # Manually add abcd_ordering_i
    df = df.sort_values('seed', axis=0)
    if rung == 1:
        df['abcd_ordering_i'] = np.array([[0, 1, 2, 3], #101
                                         [0, 1, 2, 3], #102
                                         [0, 1, 2, 3], #103
                                         [0, 1], #104
                                         [0, 1], #105
                                         [0, 1, 2, 3], #107
                                         [1, 0, 3, 2], #108
                                         [1, 2, 0, 3], #109
                                         [1, 2, 3, 0], #110
                                         [3, 1, 0, 2], #111
                                         [2, 0, 1, 3], #113
                                         [1, 0], #114
                                         [1, 3, 2, 0], #115
                                         [1, 0], #116
                                         [3, 2, 0, 1], #117
                                         [3, 1, 0, 2], #118
                                         ])
        df['H0'] = 74.151
    elif rung == 2:
        df['abcd_ordering_i'] = np.array([[0, 1, 2, 3], #119
                                         [0, 1, 2, 3], #120
                                         [0, 1, 2, 3], #121
                                         [0, 1, 2, 3], #122
                                         [0, 1, 2, 3], #123
                                         [0, 2, 1, 3], #124
                                         [0, 1], #125
                                         [0, 1], #126
                                         [3, 0, 1, 2], #127
                                         [3, 2, 0, 1], #128
                                         [3, 0, 1, 2], #129
                                         [2, 1, 0, 3], #130
                                         [3, 0, 2, 1], #131
                                         [1, 3, 2, 0], #132
                                         [1, 0], #133
                                         [0, 1], #134
                                         ])
        df['H0'] = 66.643
    else:
        raise NotImplementedError
    
    df.to_csv(save_csv_path, index=None)
    return df

def parse_closed_box(closed_box_path, row_dict=dict()):
    """Parse the lines of an open-box TDLMX text file for Rungs 0, 1, and 2

    Parameters
    ----------
    closed_box_path : str
        path to the closed box text file, `lens_info_for_Good_team.txt.txt`
    row_dict : dict
        dictionary of the row info to update. Default: dict()

    Returns
    -------
    dict
        An updated dictionary containing the information in the closed box text file

    """
    file = open(closed_box_path)
    lines = [line.rstrip('\n') for line in file]

    row_dict['z_lens'], row_dict['z_src'] = literal_eval(lines[2].split('\t')[1])
    row_dict['measured_vel_disp'] = float(lines[5].split('\t')[1].split('km/s')[0])
    row_dict['measured_vel_disp_err'] = float(lines[5].split('\t')[1].split('km/s')[1].split(':')[1])
    row_dict['measured_td'] = literal_eval(re.split(r'\(|\)', lines[7])[1])
    row_dict['measured_td_err'] = literal_eval(re.split(r'\(|\)', lines[7])[3])
    return row_dict

def parse_open_box(open_box_path, row_dict=dict()):
    """Parse the lines of an open-box TDLMX text file for Rungs 0, 1, and 2

    Parameters
    ----------
    open_box_path : str
        path to the open box text file, `lens_all_info.txt`
    row_dict : dict
        dictionary of the row info to update. Default: dict()

    Returns
    -------
    dict
        An updated dictionary containing the information in the open box text file

    """
    file = open(open_box_path)
    lines = [line.rstrip('\n') for line in file]
    row_dict['H0'] = float(re.split(r':\s|km/s/Mpc', lines[3])[-2])
    row_dict['td_distance'] = float(re.split('ls:|Mpc', lines[5])[-2])
    row_dict['time_delays'] = literal_eval(re.split(r'\(|\)', lines[7])[1])
    row_dict['lens_mass'] = literal_eval(lines[11][7:])
    row_dict['ext_shear_e1e2'], row_dict['ext_shear_bphi'] = literal_eval(re.split(r'\(|\)', lines[12])[1])
    row_dict['lens_light'] = literal_eval(lines[14].split('\t')[1])
    row_dict['host_name'] = re.split(r'\(|\)|:|\t', lines[16])[2][1:]
    row_dict['host_pos'] = literal_eval(re.split(r'\(|\)|:|\t', lines[16])[-2])
    row_dict['host_mag'] = float(re.split(r'\t|\s', lines[17])[3])
    row_dict['host_r_eff'] = float(re.split(r'\t|\s', lines[17])[7])
    row_dict['agn_src_amp'] = float(lines[20].split()[-1])
    row_dict['agn_img_pos_x'] = literal_eval(re.split(r'\(|\)', lines[21])[1])
    row_dict['agn_img_pos_y'] = literal_eval(re.split(r'\(|\)', lines[21])[3])
    row_dict['agn_img_amp'] = literal_eval(re.split(r'\(|\)', lines[22])[1])
    row_dict['host_img_mag'] = re.split('plane: |mag|', lines[23])[3]
    row_dict['agn_img_mag'] = re.split('plane: |mag|', lines[23])[7]
    row_dict['vel_disp'] = float(re.split(r'km\/s| |\t', lines[25])[1])
    row_dict['kappa_ext'] = float(lines[27].split('\t')[1])

    return row_dict


import torch
import torchvision.models as models
from torchvision.models.resnet import conv1x1, BasicBlock
import torch.nn as nn
import torch.nn.functional as F

__all__ = ['resnet34', 'resnet44', 'resnet50', 'resnet56', 'resnet101']



class BayesianBasicBlock(BasicBlock):
    """Basic block of ResNet BNN with architectural modifications from the torchvision implementation

    """
    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None, dropout_rate=0.0):
        super(BayesianBasicBlock, self).__init__(inplanes, planes, stride, downsample, groups,
                 base_width, dilation, norm_layer)
        self.dropout_rate = dropout_rate

    def forward(self, x):   
        identity = x
        out = F.dropout(x, p=self.dropout_rate)   
        out = self.conv1(out)  
        out = self.bn1(out) 
        out = self.relu(out)    

        out = F.dropout(out, p=self.dropout_rate) 
        out = self.conv2(out)   
        out = self.bn2(out) 

        if self.downsample is not None: 
            identity = self.downsample(x) 

        out += identity 
        out = self.relu(out)   
        return out

class BayesianResNet(models.ResNet):
    """ResNet BNN with architectural modifications from the torchvision implementation

    """
    def __init__(self, block, layers, num_classes=1000, zero_init_residual=False,
                 groups=1, width_per_group=64, replace_stride_with_dilation=None,
                 norm_layer=None, dropout_rate=0.0):
        self.dropout_rate = dropout_rate
        self.inplanes = 64
        super(BayesianResNet, self).__init__(block, layers, num_classes, zero_init_residual,
                 groups, width_per_group, replace_stride_with_dilation,
                 norm_layer)
        # Override first conv layer 
        self.conv1 = nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3, bias=False)
        self.include_layer4 = False if layers[-1] == 1 else True
        # If removing layer4, number of filters in FC should be 256, not 512
        if self.include_layer4:
            self._forward_impl = self._forward_impl_4layer
        else:
            self.fc = nn.Linear(256 * block.expansion, num_classes)
            self._forward_impl = self._forward_impl_3layer

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        """

        Parameters
        ----------
        block : BasicBlock class
        planes : int
            number of input filters
        blocks : int
            number of BasicBlocks for this layer (depth)

        """
        # Override _make_layer to pass in dropout_rate
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )
        # Note "layers" below represents a single layer that's returned
        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer, dropout_rate=self.dropout_rate))
        self.inplanes = planes * block.expansion
        # Note 4th layer is never used
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer, dropout_rate=self.dropout_rate))
        return nn.Sequential(*layers)

    def _forward_impl_3layer(self, x):
        # See note [TorchScript super()
        x = F.dropout(x, p=self.dropout_rate) # F not NN b/c activated during eval
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = F.dropout(x, p=self.dropout_rate)
        x = self.fc(x)
        return x

    def _forward_impl_4layer(self, x):
        # See note [TorchScript super()
        x = F.dropout(x, p=self.dropout_rate) # F not NN b/c activated during eval
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = F.dropout(x, p=self.dropout_rate)
        x = self.fc(x)
        return x

    def _forward_debug(self, x):
        # See note [TorchScript super()
        activation_map_shapes = []
        activation_map_shapes.append(x.shape)
        x = F.dropout(x, p=self.dropout_rate) # F not NN b/c activated during eval
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        activation_map_shapes.append(x.shape)
        x = self.maxpool(x)
        activation_map_shapes.append(x.shape)
        x = self.layer1(x)
        activation_map_shapes.append(x.shape)
        x = self.layer2(x)
        activation_map_shapes.append(x.shape)
        x = self.layer3(x)
        activation_map_shapes.append(x.shape)
        if self.include_layer4:
            x = self.layer4(x)
        activation_map_shapes.append(x.shape)
        x = self.avgpool(x)
        activation_map_shapes.append(x.shape)
        x = torch.flatten(x, 1)
        x = F.dropout(x, p=self.dropout_rate)
        x = self.fc(x)
        activation_map_shapes.append(x.shape)
        return activation_map_shapes

def _resnet(arch, block, layers, progress, **kwargs):
    model = BayesianResNet(block, layers, **kwargs)
    return model

def resnet34(progress=True, **kwargs):
    r"""ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet34', BayesianBasicBlock, [3, 4, 6, 3], progress,
                   **kwargs)

def resnet50(progress=True, **kwargs):
    r"""ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet50', BayesianBasicBlock, [3, 4, 6, 3], progress,
                   **kwargs)

def resnet101(progress=True, **kwargs):
    r"""ResNet-101 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet101', BayesianBasicBlock, [3, 4, 23, 3], progress,
                   **kwargs)

def resnet44(progress=True, **kwargs):
    r"""ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        progress (bool): If True, displays a progress bar of the download to stderr
    
    """
    return _resnet('resnet44', BayesianBasicBlock, [7, 7, 7, 1], progress,
                   **kwargs)

def resnet56(progress=True, **kwargs):
    r"""ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_

    Args:
        progress (bool): If True, displays a progress bar of the download to stderr
    
    """
    return _resnet('resnet56', BayesianBasicBlock, [9, 9, 9, 1], progress,
                   **kwargs)


import os
from pathlib import Path
import numpy as np
import pandas as pd
import torch
from astropy.io import fits
from torch.utils.data import Dataset
import torchvision.transforms as transforms
from baobab.data_augmentation.noise_torch import NoiseModelTorch
from baobab.sim_utils import add_g1g2_columns
from .data_utils import whiten_pixels, plus_1_log, whiten_Y_cols
import h0rton.tdlmc_utils
import h0rton.tdlmc_data

__all__ = ['TDLMCData',]

class TDLMCData(Dataset): # torch.utils.data.Dataset
    """Represents the XYData used to train or validate the BNN

    """
    def __init__(self, data_cfg, rung_i=2):
        """
        Parameters
        ----------
        dataset_dir : str or os.path object
            path to the directory containing the images and metadata
        data_cfg : dict or Dict
            copy of the `data` field of `BNNConfig`

        """
        self.__dict__ = data_cfg
        self.img_dir = os.path.join(h0rton.tdlmc_data.__path__[0], 'rung{:d}'.format(rung_i))
        self.img_paths = np.sort(list(Path(self.img_dir).rglob('*drizzled_image/lens-image.fits')))
        # Rescale pixels, stack filters, and shift/scale pixels on the fly 
        rescale = transforms.Lambda(whiten_pixels)
        log = transforms.Lambda(plus_1_log)
        transforms_list = []
        if self.log_pixels:
            transforms_list.append(log)
        if self.rescale_pixels:
            transforms_list.append(rescale)
        if len(transforms_list) == 0:
            self.X_transform = None
        else:
            self.X_transform = transforms.Compose(transforms_list)
        # Y metadata
        self.cosmo_df = h0rton.tdlmc_utils.convert_to_dataframe(rung=rung_i, save_csv_path=None)
        self.cosmo_df.sort_values('seed', axis=0, inplace=True)
        # Size of dataset
        self.n_data = self.cosmo_df.shape[0]
        # Number of predictive columns
        self.Y_dim = len(self.Y_cols)
        # Adjust exposure time relative to that used to generate the noiseless images
        self.exposure_time_factor = self.noise_kwargs.exposure_time/9600.0
        if self.add_noise:
            self.noise_model = NoiseModelTorch(**self.noise_kwargs)

    def __getitem__(self, index):
        # Image X
        img_path = self.img_paths[index]
        img = fits.getdata(img_path, ext=0)
        img *= self.exposure_time_factor
        img = img[17:-18, 17:-18] # Hacky clipping to preserve pixel scale and resize 99 x 99 to 64 x 64 
        img = torch.as_tensor(img.astype(np.float32)) # np array type must match with default tensor type
        if self.add_noise:
            img += self.noise_model.get_noise_map(img)
        img = self.X_transform(img).unsqueeze(0)
        return img

    def __len__(self):
        return self.n_data